/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   mike@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "icecryptloki91.h"

#include <QSysInfo>

extern void dump ( QByteArray* bytes );

/* 32-bit permutation function P */
/*   specifies which input bit is permuted to output bits */
/*   31 30 29 ... 2 1 0 respectively (ie in MSB to LSB order) */
char P[32] = {
	31, 23, 15, 7, 30, 22, 14, 6,
	29, 21, 13, 5, 28, 20, 12, 4,
	27, 19, 11, 3, 26, 18, 10, 2,
	25, 17, 9, 1, 24, 16, 8, 0
};

/*
*      sfn_desc - a desriptor table specifying which irreducible polynomial
*              and exponent is to be used for each of the 16 S functions in
*              the Loki S-box
*/
typedef struct
{
	short   gen;            /* irreducible polynomial used in this field */
	short   exp;            /* exponent used to generate this s function */
}
sfn_desc;

/*
*      sfn - the table specifying the irreducible polys & exponents used
*              in the Loki S-boxes for the Loki algorithm
*/
sfn_desc sfn[] = {
	{ /* 101110111 */ 375, 31},
	{ /* 101111011 */ 379, 31},
	{ /* 110000111 */ 391, 31},
	{ /* 110001011 */ 395, 31},
	{ /* 110001101 */ 397, 31},
	{ /* 110011111 */ 415, 31},
	{ /* 110100011 */ 419, 31},
	{ /* 110101001 */ 425, 31},
	{ /* 110110001 */ 433, 31},
	{ /* 110111101 */ 445, 31},
	{ /* 111000011 */ 451, 31},
	{ /* 111001111 */ 463, 31},
	{ /* 111010111 */ 471, 31},
	{ /* 111011101 */ 477, 31},
	{ /* 111100111 */ 487, 31},
	{ /* 111110011 */ 499, 31},
	{ 00, 00}
};

/*
*	ROL12(b) - macro to rotate 32-bit block b left by 12 bits
*	ROL13(b) - macro to rotate 32-bit block b left by 13 bits
*/
#define ROL12(b) b = ((b << 12) | (b >> 20));
#define ROL13(b) b = ((b << 13) | (b >> 19));


ICECryptLOKI91::ICECryptLOKI91()
: ICECrypt()
, last_xdatal ( 0xCACACACA )
, last_xdatar ( 0xCACACACA )
, cbc ( true )
{}


ICECryptLOKI91::~ICECryptLOKI91()
{}

void ICECryptLOKI91::cipher ( bool encrypt, QString key, QByteArray* in, QByteArray* out )
{
	//fprintf( stderr, "%s: key='%s'\n", encrypt?"encrypt":"decrypt", key.toAscii().data() );
	//fprintf( stderr, "%s(I): in size=%d, out size=%d\n", encrypt?"encrypt":"decrypt", in->size(), out->size() );
	//dump( in );
	if ( !mInitialized )
	{
		init ( ( unsigned char* ) key.toLatin1().data() );
		mInitialized = true;
	}
	out->resize ( in->size() );
	for ( int index = 0; index < in->size(); index+=8 )
	{
		unsigned char cipher_block[8];

		/*
		** Block Padding...
		*/
		quint32 remaining=in->size()-index;
		if ( remaining < 8 )
		{
			int pad = ( 8-remaining );
			//fprintf( stderr, "pad=%d\n", pad );
			in->resize ( in->size() + pad );
			out->resize ( out->size() + pad );
		}

		/*
		** cipher...
		*/
		memcpy ( cipher_block,& ( ( in->data() ) [index] ),8 );
		if ( encrypt )
		{
			encipher ( cipher_block );
		}
		else
		{
			decipher ( cipher_block );
		}
		memcpy ( & ( ( out->data() ) [index] ),cipher_block,8 );

	}
	//fprintf( stderr, "%s(O): in size=%d, out size=%d\n", encrypt?"encrypt":"decrypt", in->size(), out->size() );
	//dump( out );
}


	/*
	*      bswap(b) - exchanged bytes in each longword of a 64-bit data block
	*                      on little-endian machines where byte order is reversed
	*/

void ICECryptLOKI91::bswap ( unsigned char *cb )
{
	if ( QSysInfo::ByteOrder == QSysInfo::LittleEndian )
	{
		char   c;
		c = cb[0]; cb[0] = cb[3]; cb[3] = c;
		c = cb[1]; cb[1] = cb[2]; cb[2] = c;
		c = cb[4]; cb[4] = cb[7]; cb[7] = c;
		c = cb[5]; cb[5] = cb[6]; cb[6] = c;
	}
}

	/*
	*	setlokikey(key) - save 64-bit key for use in encryptions & decryptions
	*			  and compute sub-keys using the key schedule
	*/

void ICECryptLOKI91::init ( unsigned char* key )
{
	int	i;
	quint32	KL, KR;

	bswap ( key );			/* swap bytes round if little-endian */

	KL = ( ( quint32 * ) key ) [0];
	KR = ( ( quint32 * ) key ) [1];

	for ( i=0; i<ROUNDS; i+=4 )
	{	/* Generate the 16 subkeys */
		loki_subkeys[i] = KL;
		ROL12 ( KL );
		loki_subkeys[i+1] = KL;
		ROL13 ( KL );
		loki_subkeys[i+2] = KR;
		ROL12 ( KR );
		loki_subkeys[i+3] = KR;
		ROL13 ( KR );
	}

	bswap ( key );			/* swap bytes back if little-endian */
}


/*
*	enloki(b) - main LOKI91 encryption routine, this routine encrypts one
*		64-bit block b using the LOKI91 algorithm with loki_subkeys
*
*		nb: The 64-bit block is passed as two longwords. For the
*			purposes of the LOKI89 algorithm, the bits are numbered:
*			    [63 62 .. 33 32] [31 30 ... 1 0]
*			The L (left) half is b[0], the R (right) half is b[1]
*
*/

void ICECryptLOKI91::encipher ( unsigned char* b )
{
	int	i;
	quint32	L, R;		/* left & right data halves  */

	bswap ( b );			/* swap bytes round if little-endian */

	L = ( ( quint32 * ) b ) [0];
	R = ( ( quint32 * ) b ) [1];

	for ( i=0; i<ROUNDS; i+=2 )
	{	/* Encrypt with the 16 subkeys */
		L ^= f ( R, loki_subkeys[i] );
		R ^= f ( L, loki_subkeys[i+1] );
	}

	( ( quint32 * ) b ) [0] = R;		/* Y = swap(LR) */
	( ( quint32 * ) b ) [1] = L;

	bswap ( b );			/* swap bytes round if little-endian */
}

	
/*
*	deloki(b) - main LOKI91 decryption routine, this routine decrypts one
*		64-bit block b using the LOKI91 algorithm with loki_subkeys
*
*		Decryption uses the same algorithm as encryption, except that
*		the subkeys are used in reverse order.
*/
void ICECryptLOKI91::decipher ( unsigned char *b )
{
	int	i;
	quint32	L, R;			/* left & right data halves  */

	bswap ( b );			/* swap bytes round if little-endian */

	L = ( ( quint32 * ) b ) [0];			/* LR = X XOR K */
	R = ( ( quint32 * ) b ) [1];

	for ( i=ROUNDS; i>0; i-=2 )
	{		/* subkeys in reverse order */
		L ^= f ( R, loki_subkeys[i-1] );
		R ^= f ( L, loki_subkeys[i-2] );
	}

	( ( quint32 * ) b ) [0] = R;			/* Y = LR XOR K */
	( ( quint32 * ) b ) [1] = L;

	bswap ( b );			/* swap bytes round if little-endian */
}

	
/*
*	f(r, k) - is the complex non-linear LOKI function, whose output
*		is a complex function of both input data and sub-key.
*
*	The data is XORed with the subkey, then expanded into 4 x 12-bit
*	values, which are fed into the S-boxes. The 4 x 8-bit outputs
*	from the S-boxes are permuted together to form the 32-bit value
*	which is returned.
*
*	In this implementation the outputs from the S-boxes have been
*	pre-permuted and stored in lookup tables.
*/

#define MASK12	0x0fff			/* 12 bit mask for expansion E */

quint32 ICECryptLOKI91::f ( quint32 r, quint32 k )
{
	quint32	a, b, c;		/* 32 bit S-box output, & P output */

	a = r ^ k;			/* A = R(i-1) XOR K(i) */

	/* want to use slow speed/small size version */
	b = ( ( quint32 ) s ( ( a         & MASK12 ) ) ) | /* B = S(E(R(i-1))^K(i)) */
		( ( quint32 ) s ( ( ( a >>  8 ) & MASK12 ) ) <<  8 ) |
		( ( quint32 ) s ( ( ( a >> 16 ) & MASK12 ) ) << 16 ) |
		( ( quint32 ) s ( ( ( ( a >> 24 ) | ( a << 8 ) ) & MASK12 ) ) << 24 );

	perm32 ( &c, &b, P );		/* C = P(S( E(R(i-1)) XOR K(i))) */

	return c;			/* f returns the result C */
}

/*
*	s(i) - return S-box value for input i
*/
qint16 ICECryptLOKI91::s ( quint32 i )
{
	qint16	r, c, v, t;

	r = ( ( i>>8 ) & 0xc ) | ( i & 0x3 );		/* row value-top 2 & bottom 2 */
	c = ( i>>2 ) & 0xff;			/* column value-middle 8 bits */
	t = ( c + ( ( r * 17 ) ^ 0xff ) ) & 0xff;	/* base value for Sfn */
	v = exp8 ( t, sfn[r].exp, sfn[r].gen );	/* Sfn[r] = t ^ exp mod gen */
	return v;
}

/*
*	perm32(out, in, perm) is the general permutation of a 32-bit input
*		block to a 32-bit output block, under the control of a
*		permutation array perm. Each element of perm specifies which
*		input bit is to be permuted to the output bit with the same
*		index as the array element.
*
*	nb: to set bits in the output word, as mask with a single 1 in it is
*		used. On each step, the 1 is shifted into the next location
*/

#define	MSB	0x80000000L		/* MSB of 32-bit word */

void ICECryptLOKI91::perm32 ( quint32 *out, quint32 *in, char *perm )
{
	quint32	mask = MSB;		/* mask used to set bit in output    */
	int	i, o, b;	/* input bit no, output bit no, value */
	char *p = perm;	/* ptr to permutation array  */

	*out = 0;			/* clear output block */
	for ( o=0; o<32; o++ )
	{		/* For each output bit position o */
		i = ( int ) *p++;		/* get input bit permuted to output o */
		b = ( *in >> i ) & 01;	/* value of input bit i */
		if ( b )			/* If the input bit i is set */
			*out |= mask;		/*  OR in mask to output i */
		mask >>= 1;			/* Shift mask to next bit    */
	}
}

	
/*
*	mult8(a, b, gen) - returns the product of two binary
*		strings a and b using the generator gen as the modulus
*			mult = a * b mod gen
*		gen generates a suitable Galois field in GF(2^8)
*/

#define SIZE 256		/* 256 elements in GF(2^8) */

qint16 ICECryptLOKI91::mult8 ( qint16 a, qint16 b, qint16 gen )
{
	qint16	product = 0;		/* result of multiplication */

	while ( b != 0 )
	{			/* while multiplier is non-zero */
		if ( b & 01 )
			product ^= a;	/*   add multiplicand if LSB of b set */
		a <<= 1;		/*   shift multiplicand one place */
		if ( a >= SIZE )
			a ^= gen;	/*   and modulo reduce if needed */
		b >>= 1;		/*   shift multiplier one place  */
	}
	return product;
}

/*
*	exp8(base, exponent, gen) - returns the result of
*		exponentiation given the base, exponent, generator of GF,
*			exp = base ^ exp mod gen
*/

qint16 ICECryptLOKI91::exp8 ( qint16 base, qint16 exponent, qint16 gen )
{
	qint16	accum = base;							/* superincreasing sequence of base */
	qint16	result = 1;								/* result of exponentiation	    */

	if ( base == 0 )								/* if zero base specified then      */
		return 0;									/* the result is "0" if base = 0    */

	while ( exponent != 0 )							/* repeat while exponent non-zero */
	{	
		if ( ( exponent & 0x0001 ) == 0x0001 )		/* multiply if exp 1 */
			result = mult8 ( result, accum, gen );
		exponent >>= 1;								/* shift exponent to next digit */
		accum = mult8 ( accum, accum, gen );		/* & square  */
	}
	return result;
}
