/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "icesql.h"

#include <QDir>
#include <QFile>
#include <QString>
#include <QFileInfoList>
#include <QFileInfo>

ICESql* ICESql::mInstance=NULL;
QList<ICESqlItem*>	ICESql::mItemList;

ICESql::ICESql(QObject *parent)
 : QObject(parent)
{
	if ( mInstance == NULL ) {
		mInstance = this;
	} else {
		emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::ICESql(): Only one instance allowed! Deleting..." );
		delete this; /** this should be fun! */
	}
}


ICESql::~ICESql()
{
	for( int index=0; index < mItemList.count(); index++ ) {
		delete mItemList.at( index );
	}
	if ( mInstance == this ) {
		mInstance = NULL;
	}
}


/**
 ** \brief find an sql query by name, and return the text of the query.
 ** \param name The name of the query.
 ** \return Returns the text of the SQL query, or an empty string if not found.
 */
QString ICESql::find( QString type, QString name ) 
{
	QString rc;
	for( int index = 0; index < mItemList.count(); index++ )
	{
		if ( mItemList.at( index )->type() == type && mItemList.at( index )->name() == name ) {
			return mItemList.at( index )->text();
		}
	}
	emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::find(\"" + type + "\",\"" + name + "\"): Not found!" );
	return rc;
}


/**
 ** \brief Perform an SQL query for all queries of type.
 ** \param type The type of the query.
 ** \return Returns the result of the query, aborts on first error and returns query result.
 */
QSqlQuery ICESql::queryAll( QString type )
{	
	QSqlQuery rc;
	QString name;
	for( int index = 0; index < mItemList.count() && rc.lastError().type() ==  QSqlError::NoError; index++ )
	{
		if ( mItemList.at( index )->type() == type ) {
			QString name = mItemList.at( index )->name();
			rc = query( type, name );
			if ( rc.lastError().type() != QSqlError::NoError ) {	
				emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::queryAll(\"" + type + "\",\"" + name + "\"): " + rc.lastError().text() );
			} else {
				emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::queryAll(\"" + type + "\",\"" + name + "\")" );
			}
		}
	}
	return rc;
}

/**
** \brief Perform an SQL query, performing substitutions in the query text.
** \param name The name of the query.
** \param ... Variable arguments to substitute in the SQL text.
** \return Returns the result of the query.
*/
QSqlQuery ICESql::query( QString type, QString name, ... )
{
	va_list ap;
	va_start(ap,name);
	QSqlQuery rc;
    QString text = find( type, name );
	QString queryText;
	for ( int index=0; index < text.length(); index++ ) {
		if ( text[ index ] == '%' ) {
			if ( ++index < text.length() ) {
				switch( text[ index ].toAscii() ) {
					case '%':
						queryText += '%';
						break;
					case 'L':		/** quint64 (unsigned) */
					{
						QString str = QString::number( (quint64)va_arg(ap,quint64) );
						queryText += str;
					}
					break;
					case 'l':		/** qint64 (signed) */
					{
						QString str = QString::number( (qint64)va_arg(ap,qint64) );
						queryText += str;
					}
					break;
					case 'c':		/** signed char */
					{
						QChar ch( (char)va_arg(ap,int) );
						queryText += ch;
					}
					break;
					case 's':		/** asciiz string pointer*/
					{
						QString str( (char*)va_arg(ap,char*) );
						queryText += str;
					}
					break;
					case 'S':		/** Unicode QString pointer */
					{
						QString str( *((QString*)va_arg(ap,QString*)) );
						queryText += str;
					}
					break;
					case 'f':
					{
						QString str = QString::number( (double)va_arg(ap,double) );
						queryText += str;
					}
					break;
					case 'b':
					{
						bool b = (bool)va_arg(ap,int);
						QString str = b?"true":"false";
						queryText += str;
					}
					break;
					default:
						emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::query(\"" + type + "\",\"" + name + "\",...): No format char '" + text[index] + "'!" );
					break;
				}
			} 
		} else {
			queryText += text[ index ];
		}
	}
	if ( queryText.length() ) {
		rc = QSqlQuery( queryText );
		if ( rc.lastError().type() != QSqlError::NoError ) {	
			emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::query(\"" + type + "\",\"" + name + "\"): " + rc.lastError().text() );
		}
	} else {
		emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::query(\"" + type + "\",\"" + name + "\",...): No query text!" );
	}
	return rc;
}

/**
** \brief Load the sql queries from the directory specified by path.
** \param path The path in which to locate the .sql files.
** \return true on success or false on failure.
*/
bool ICESql::load( QString path )
{
	bool rc=true;
	QStringList filters;
	filters << "*";
	QDir dir( path );
	if ( dir.exists() ) {
		QString type = dir.dirName();
		QFileInfoList fileInfos = dir.entryInfoList( filters, (QDir::Readable | QDir::Files | QDir::Dirs | QDir::NoDotAndDotDot), QDir::Name );
		for( int index=0; rc && index < fileInfos.count(); index++ ) {
			QFileInfo fi =  fileInfos[ index ];
			if ( fi.isFile() && ( fi.suffix() == "sql" || fi.suffix() == "SQL" ) ) {
				QFile file( fi.absoluteFilePath() );
				if ( file.open( QIODevice::ReadOnly | QIODevice::Text ) ) {
					QTextStream in ( &file );
					QString text;
					while ( !in.atEnd() ) {
						QString line = in.readLine().simplified();
						if ( line.length() && line.left(2) != "--" ) {
							text += line;
						}
					}
					QString itemName = fi.completeBaseName();
					ICESqlItem* sqlItem = new ICESqlItem( type, itemName, text );
					mItemList.append( sqlItem );
					file.close();
				} else {
					emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::load(\"" + path + "\"): Cannot open!" );
					rc = false;
				}
			} else if ( fi.isDir() ) {
				rc = load( fi.absoluteFilePath() );
			} else {
				emit mInstance->statusMessage( mInstance, ICESession::ReasonDatabase, "ICESql::load(\"" + path + "\"): skipping: " + fi.absoluteFilePath() );
			}
		}
	} else {
		rc = false;
	}
	return rc;	
}

