/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "icepacket.h"

#include <QDataStream>

extern void dump( QByteArray* bytes );

#define PROTOCOL_VERSION 1.1

ICEPacket::ICEPacket()
: QObject(NULL)
, mType(Empty)
, mCode(CodeOK)
, mObjectType(0)
, mCipher(CipherXOR)
, mProtocolVersion(PROTOCOL_VERSION)
, mPlatform(Unknown)
, mInteger(0)
{
}

ICEPacket::ICEPacket(ICEPacket& other)
: QObject(NULL)
{
	copy(other);
}

ICEPacket::~ICEPacket()
{
}

/**
	\brief Deep Copy from other object.

	\param other Other packet to copy from.
 **/
ICEPacket& ICEPacket::operator=(ICEPacket& other)
{
	copy(other);
	return *this;
}

/**
	\brief Deep Copy from other object.

	\param other Other packet to copy from.
 **/
void ICEPacket::copy(ICEPacket& other)
{
	setType             ( other.getType() );
	setCode             ( other.getCode() );
	setLogin            ( other.getLogin() );
	setObjectId			( other.getObjectId() );
	setObjectType		( other.getObjectType() );
	setKey              ( other.getKey() );
	setCipher           ( other.getCipher() );
	setProtocolVersion	( other.getProtocolVersion() );
	setPlatform         ( other.getPlatform() );
	setString			( other.getString() );
	setInteger			( other.getInteger() );
	setLocation			( other.getLocation() );
	setData				( other.getData() );
	setUi				( other.getUi() );
	setUiObjects		( other.getUiObjects() );
	setEvent			( other.getEvent() );
}


/**
	\brief Encode "this" using "crypt" cypher to produce a ciphered QByteArray.

	\param crypt Cypher to use for encoding.

	\return QByteArray - The cypher text.
 **/
QByteArray& ICEPacket::encode(ICECrypt* crypt)
{
	if ( getType() != Empty ) {
		mClearText.clear();
		mCipherText.clear();
		QDataStream stream( &mClearText, QIODevice::Append );
		stream << (qint32)getType();
		stream << (qint32)mCode;
		switch( getType() ) {
			case Empty:
				break;
			case Ping:
				break;
			case Fault:
				break;

			case RequestConnection:
				stream << mProtocolVersion;
				stream << mLogin;
				{
					quint32 iplatform = (quint32)mPlatform;
					stream << iplatform;
				}
				break;
			case RequestDisconnect:
				break;
			case RequestConnectNewLoginID:
				stream << mLogin;
				stream << mKey;
				break;
			case RequestLogin:
				stream << (qint32)mCipher;
				break;
			case RequestDeleteLogin:
				break;
			case RequestSession:
				stream << mData;
				break;
			case RequestPasswordChange:
				stream << mKey;
				break;
			case RequestSessionObjectConstruct:
				stream << mObjectType;
				stream << mObjectId;
				stream << mLocation;
				stream << mUi;
				break;
			case RequestSessionObjectDestroy:
				stream << mObjectId;
				break;
			case RequestSessionObjectIncrementReferences:
				stream << mObjectId;
				break;	
			case RequestSessionObjectDecrementReferences:
				stream << mObjectId;
				break;
			case RequestSessionObjectUi:
				stream << mObjectId;
				stream << mUi;
				break;
			case RequestSessionObjectUiObjects:
				stream << mObjectId;
				stream << mUiObjects;
				break;
			case RequestSessionObjectEvent:
				stream << mObjectId;
				stream << mEvent;
				break;
			
				
			case ReplyConnectNewLoginID:
			case ReplyConnection:
			case ReplyDisconnect:
				break;
			case ReplyLogin:
				stream << mData;
				break;
			case ReplyDeleteLogin:
			case ReplySession:
			case ReplyPasswordChange:
				break;
			case ReplySessionObjectConstruct:
				stream << mObjectId;
				stream << mLocation;
				stream << mUi;
				break;
			case ReplySessionObjectDestroy:
			case ReplySessionObjectIncrementReferences:
			case ReplySessionObjectDecrementReferences:
				stream << mObjectId;
				break;
			case ReplySessionObjectUi:
				stream << mObjectId;
				stream << mUi;
				break;
			case ReplySessionObjectUiObjects:
				stream << mObjectId;
				stream << mUiObjects;
				break;
			case ReplySessionObjectEvent:
				stream << mObjectId;
				break;
				
			case AsyncPopupMessage:
				stream << mString;
				break;

			default:
				emit packetError((mCode=CodeBadPacketType),QString::number(getType()));
				break;
		}
		stream.device()->close();
		quint32 iheader;
		unsigned char header[4];
		if ( crypt != NULL ) {
			crypt->cipher( true, getKey(), &mClearText, &mCipherText );
			iheader = mCipherText.size();
			//fprintf( stderr, "packet encode: %d bytes.\n", iheader );
			header[0] = (unsigned char)((iheader&0xFF000000)>>24);
			header[1] = (unsigned char)((iheader&0x00FF0000)>>16);
			header[2] = (unsigned char)((iheader&0x0000FF00)>>8);
			header[3] = (unsigned char)((iheader&0x000000FF));
			mCipherText.prepend( QByteArray( (char*)header, 4 ) );
			return mCipherText;
		} else {
			iheader = mClearText.size();
			//fprintf( stderr, "packet encode: %d bytes.\n", iheader );
			header[0] = (unsigned char)((iheader&0xFF000000)>>24);
			header[1] = (unsigned char)((iheader&0x00FF0000)>>16);
			header[2] = (unsigned char)((iheader&0x0000FF00)>>8);
			header[3] = (unsigned char)((iheader&0x000000FF));
			mClearText.prepend( QByteArray( (char*)header, 4 ) );
			return mClearText;
		}
	}
	return mEmptyText;
}

/**
	\brief Decode input stream "io" into "this", using "crypt" cypher.

	\param io I/O Device to get input stream from..
	\param crypt Cypher to use for encoding.

	\return void
 **/
void ICEPacket::decode(QIODevice* io, ICECrypt* crypt)
{
	qint32 icode=0;
	quint32 iheader=0;
	unsigned char header[4];
	qint64 rc;
	char* pData;

	mClearText.clear();
	mCipherText.clear();
	io->read( (char*)header, 4 );
	iheader |= ((header[0] << 24) & 0xFF000000);
	iheader |= ((header[1] << 16) & 0x00FF0000);
	iheader |= ((header[2] << 8)  & 0x0000FF00);
	iheader |= ((header[3])       & 0x000000FF);
	//fprintf( stderr, "packet decode: %d bytes.\n", iheader );
	mCipherText.resize( iheader );
	pData = mCipherText.data();
	while ( (rc = io->read( pData, iheader )) <= 0 ) {
		if ( rc > 0 ) {
			pData = &pData[rc];
			iheader -= rc;
		} else {
			emit packetError((mCode=CodeIOReadError),QString::number(rc));
			mCipherText.clear();
			break;
		}
	}
	if ( mCipherText.size() > 0 ) {
		if ( crypt != NULL ) {
			crypt->cipher( false, getKey(), &mCipherText, &mClearText );
		} else {
			mClearText = mCipherText;
		}
		mCipherText.clear();
		QDataStream stream( &mClearText, QIODevice::ReadOnly );
		qint32 itype;
		stream >> itype;
		stream >> icode;
		mType = (PacketType)itype;
		switch( getType() ) {
			case Empty:                     break;
			case Ping:                      break;
			case Fault:                     break;

			case RequestConnection:
				stream >> mProtocolVersion;
				stream >> mLogin;
				{
					quint32 iplatform;
					stream >> iplatform;
					mPlatform = (PlatformType)iplatform;
				}
				break;
			case RequestDisconnect:
				break;
			case RequestConnectNewLoginID:
				stream >> mLogin;
				stream >> mKey;
				break;
			case RequestLogin:
			{
				int icypher;
				stream >> icypher;
				mCipher = (PacketCipher)icypher;
			}
			break;
			case RequestDeleteLogin:
				break;
			case RequestSession:
				stream >> mData;
				break;
			case RequestPasswordChange:
				stream >> mKey;
				break;

			case RequestSessionObjectConstruct:
				stream >> mObjectType;
				stream >> mObjectId;
				stream >> mLocation;
				stream >> mUi;
				break;
			case RequestSessionObjectDestroy:
				stream >> mObjectId;
				break;	
			case RequestSessionObjectIncrementReferences:
				stream >> mObjectId;
				break;
			case RequestSessionObjectDecrementReferences:
				stream >> mObjectId;
				break;
			case RequestSessionObjectUi:
				stream >> mObjectId;
				stream >> mUi;
				break;
			case RequestSessionObjectUiObjects:
				stream >> mObjectId;
				stream >> mUiObjects;
				break;
			case RequestSessionObjectEvent:
				stream >> mObjectId;
				stream >> mEvent;
				break;
	 
			case ReplyConnectNewLoginID:
			case ReplyConnection:
			case ReplyDisconnect:
				break;
			case ReplyLogin:
				stream >> mData;
				break;
			case ReplyDeleteLogin:
			case ReplySession:
			case ReplyPasswordChange:
				break;
			case ReplySessionObjectConstruct:
				stream >> mObjectId;
				stream >> mLocation;
				stream >> mUi;
				break;
			case ReplySessionObjectDestroy:
			case ReplySessionObjectIncrementReferences:
			case ReplySessionObjectDecrementReferences:
				stream >> mObjectId;
				break;
			case ReplySessionObjectUi:
				stream >> mObjectId;
				stream >> mUi;
				break;
			case ReplySessionObjectUiObjects:
				stream >> mObjectId;
				stream >> mUiObjects;
				break;
			case ReplySessionObjectEvent:
				stream >> mObjectId;
				break;

			case AsyncPopupMessage:
				stream >> mString;
				break;

			default:
				emit packetError((mCode=CodeBadPacketType),QString::number(getType()));
				break;
		}
		mCode = (PacketCode)icode;
		stream.device()->close();
	}
}

/**
    \brief Information about the packet erturned as a string suitable for logging.
 **/
QString ICEPacket::toString()
{
	QString msg;
	switch( getType() ) {

		case ICEPacket::Empty:
			msg += "Empty"; break;
		case ICEPacket::Ping:
			msg += "Ping"; break;
		case ICEPacket::Fault:
			msg += "Fault"; break;

		case ICEPacket::RequestConnection:
			msg += "RequestConnection"; break;
		case ICEPacket::RequestDisconnect:
			msg += "RequestDisconnect"; break;
		case ICEPacket::RequestConnectNewLoginID:
			msg += "RequestConnectNewLoginID"; break;
		case ICEPacket::RequestLogin:
			msg += "RequestLogin"; break;
		case ICEPacket::RequestDeleteLogin:
			msg += "RequestDeleteLogin"; break;
		case ICEPacket::RequestSession:
			msg += "RequestSession"; break;
		case ICEPacket::RequestPasswordChange:
			msg += "RequestPasswordChange"; break;
		case ICEPacket::RequestSessionObjectConstruct:
			msg += "RequestSessionObjectConstruct"; break;
		case ICEPacket::RequestSessionObjectDestroy:
			msg += "RequestSessionObjectDestroy"; break;
		case ICEPacket::RequestSessionObjectIncrementReferences:
			msg += "RequestSessionObjectIncrementReferences"; break;
		case ICEPacket::RequestSessionObjectDecrementReferences:
			msg += "RequestSessionObjectDecrementReferences"; break;
		case ICEPacket::RequestSessionObjectUiObjects:
			msg += "RequestSessionObjectUiObjects"; break;
		case ICEPacket::RequestSessionObjectUi:
			msg += "RequestSessionObjectUi"; break;
		case ICEPacket::RequestSessionObjectEvent:
			msg += "RequestSessionObjectEvent"; break;
	 
		case ICEPacket::ReplyConnection:
			msg += "ReplyConnection"; break;
		case ICEPacket::ReplyDisconnect:
			msg += "ReplyDisconnect"; break;
		case ICEPacket::ReplyConnectNewLoginID:
			msg += "ReplyNewLiginID"; break;
		case ICEPacket::ReplyLogin:
			msg += "ReplyLogin"; break;
		case ICEPacket::ReplyDeleteLogin:
			msg += "ReplyDeleteLogin"; break;
		case ICEPacket::ReplySession:
			msg += "ReplySession"; break;
		case ICEPacket::ReplyPasswordChange:
			msg += "ReplyPasswordChange"; break;
		case ICEPacket::ReplySessionObjectConstruct:
			msg += "ReplySessionObjectConstruct"; break;
		case ICEPacket::ReplySessionObjectDestroy:
			msg += "ReplySessionObjectDestroy"; break;
		case ICEPacket::ReplySessionObjectIncrementReferences:
			msg += "ReplySessionObjectIncrementReferences"; break;
		case ICEPacket::ReplySessionObjectDecrementReferences:
			msg += "ReplySessionObjectDecrementReferences"; break;
		case ICEPacket::ReplySessionObjectUiObjects:
			msg += "ReplySessionObjectUiObjects"; break;
		case ICEPacket::ReplySessionObjectUi:
			msg += "ReplySessionObjectUi"; break;
		case ICEPacket::ReplySessionObjectEvent:
			msg += "ReplySessionObjectEvent"; break;

       
		case ICEPacket::AsyncPopupMessage:
			msg += "AsyncPopupMessage"; break;

		default:
			msg += "Unknown Packet Type (" + QString::number(getType()) + ")"; break;
	}
    
	msg += "|rc(" + QString::number(getCode() ) + ")|";
	
	switch ( getCode() ) {
		case CodeOK: 						msg += tr( "OK" ); break;

		case CodeProtocolError:				msg += tr( "Protocol Error" ); break;
		case CodeTooBusy:					msg += tr( "Too Busy" ); break;
		case CodeDuplicateLoginID:			msg += tr( "Duplicate Login ID" ); break;
		case CodeSuspendedLoginID:			msg += tr( "Suspended Login ID" ); break;
		case CodeUnknownLoginID: 			msg += tr( "Unknown Login ID" ); break;
		case CodePasswordTooWeak:			msg += tr( "Password Too Weak" ); break;
		case CodeBlockedNetAddress:			msg += tr( "Blocked Network Address" ); break;
		case CodeFailedChallenge:			msg += tr( "Failed Challenge" ); break;
		case CodeCipherNotSupported:		msg += tr( "Cipher Not Supported" ); break;
		case CodeCreateNewIDFailed:			msg += tr( "Create New ID Failed" ); break;
		case CodeTooManyNewIDsIssued: 		msg += tr( "Too Many New IDs Issued" ); break;
		case CodeCreateNewIDNotAllowed:		msg += tr( "Create New ID Not Allowed" ); break;
		case CodeInvalidCredentials: 		msg += tr( "Invalid Credentials" ); break;
							
		case CodeDirectoryEmpty: 			msg += tr( "Directory Empty" ); break;
		case CodeInsufficientAccess:		msg += tr( "Insufficient Access" ); break;
		case CodeStorageLimitExceeded:		msg += tr( "Storage Limit Exceeded" ); break;
		case CodeUpdateFailed:				msg += tr( "Update Failed" ); break;
		case CodeInsertFailed:				msg += tr( "Insert Failed" ); break;
		case CodeFileNotFound:				msg += tr( "File Not Found" ); break;
		case CodeDirectoryNotFound:			msg += tr( "Directory Not Found" ); break;
		case CodeDirectoryExists:			msg += tr( "Directory Exists" ); break;
		case CodeFileExists:				msg += tr( "File Exists" ); break;
		case CodeFileTypeMismatch:			msg += tr( "File Type Mismatch" ); break;
		case CodeFileUnknownMode:			msg += tr( "File Unknown Mode" ); break;
		case CodeFileNotOpened:				msg += tr( "File Not Opened" ); break;
		case CodeFileOpenFailed:			msg += tr( "File Open Failed" ); break;
		case CodeFileSeekFailed:			msg += tr( "File Seek Failed" ); break;
		case CodeFileReadFailed: 			msg += tr( "File Read Failed" ); break;
		case CodeFileWriteFailed:			msg += tr( "File Write Failed" ); break;
		case CodeFileCreateFailed:			msg += tr( "File Create Failed" ); break;
		case CodeFileEOF:					msg += tr( "File EOF Reached" ); break;
										
		case CodeBadPacketType: 			msg += tr( "Bad Packet Type" ); break;
		case CodeIOReadError:				msg += tr( "I/O Read Error" ); break;
		default:							msg += tr( "Unknown Code" ); break;					
	}
	
	return msg; 
}
