/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ICEPACKET_H
#define ICEPACKET_H

#include <QObject>
#include <QObject>
#include <QByteArray>
#include <QString>
#include <QDateTime>
#include <QIODevice>
#include <QTcpSocket>
#include <QList>
#include <QDataStream>
#include <QBuffer>
#include <QStringList>

#include "icecrypt.h"
#include "iceevent.h"

/**
@author Michael Sharkey
	Serialize and de-serialize packet data using a cipher.
 */

#define DirListReturnID				1001
#define	IOQueueReturnID				1002

class ICEPacket : public QObject
{
	Q_OBJECT
	public:
		/**
		* Operating system / platform used to determine method of
		* remote to local path translation.
		*/
		typedef enum {
			Unknown=0,
			Unix,
			Linux,
			Qtopia,
			Windows,
			MacOSX
		} PlatformType;
        
		/**
		 ** Packet types describe what type of data to expect
		 ** in the packet.
		 */
		typedef enum {
			Empty=0,
			Ping=1,
			Fault=2,
			
			RequestConnection=100,
			RequestDisconnect,
			RequestConnectNewLoginID,
			RequestLogin,
			RequestDeleteLogin,
			RequestSession,
			RequestPasswordChange,
			
			ReplyConnection=200,
			ReplyDisconnect,
			ReplyConnectNewLoginID,
			ReplyLogin,
			ReplyDeleteLogin,
			ReplySession,
			ReplyPasswordChange,

   			RequestSessionObjectConstruct=300,
			RequestSessionObjectDestroy,
	        RequestSessionObjectIncrementReferences,
		 	RequestSessionObjectDecrementReferences,
			RequestSessionObjectUi,
			RequestSessionObjectUiObjects,
			RequestSessionObjectLocation,
			RequestSessionObjectEvent,

			ReplySessionObjectConstruct=600,
			ReplySessionObjectDestroy,
			ReplySessionObjectIncrementReferences,
			ReplySessionObjectDecrementReferences,
			ReplySessionObjectUi,
			ReplySessionObjectUiObjects,
			ReplySessionObjectLocation,
			ReplySessionObjectEvent,

			AsyncPopupMessage=1000,

		} PacketType;

		/**
		 ** Each packet carries exactly one of these codes.
		 */
		typedef enum {
			CodeOK=0,

			CodeProtocolError=100,
			CodeTooBusy,
			CodeDuplicateLoginID,
			CodeSuspendedLoginID,
			CodeUnknownLoginID,
			CodePasswordTooWeak,
			CodeBlockedNetAddress,
			CodeFailedChallenge,
			CodeCipherNotSupported,
			CodeCreateNewIDFailed,
			CodeTooManyNewIDsIssued,
			CodeCreateNewIDNotAllowed,
			CodeInvalidCredentials,
			
			CodeDirectoryEmpty=300,
   			CodeInsufficientAccess,
	  		CodeStorageLimitExceeded,
			CodeUpdateFailed,
			CodeInsertFailed,
			CodeFileNotFound,
			CodeDirectoryNotFound,
			CodeDirectoryExists,
			CodeFileExists,
			CodeFileTypeMismatch,
			CodeFileUnknownMode,
			CodeFileNotOpened,
			CodeFileOpenFailed,
			CodeFileSeekFailed,
			CodeFileReadFailed,
			CodeFileWriteFailed,
			CodeFileCreateFailed,
   			CodeFileEOF,
	  
			CodeBadPacketType=1000,
			CodeIOReadError,

   			CodeUnknownEventType=5000,
   
		} PacketCode;

		/**
		 ** Types of cyphers the client can request
		 ** during connection negotiation
		 */
		typedef enum {
			CipherXOR=0,
			CipherBlowfish,
			CipherSEAL,
			CipherDES,
			CipherRC5,
			CipherLOKI91,
			Cipher3Way,
			CipherGOST,
			CipherIDEA,
			CipherA5,
			CipherPIKE,
			CipherMAX,   /* always leave this at the end of list */
		} PacketCipher;

		static const quint32 Read=0x01;
		static const quint32 Write=0x02;

		ICEPacket();
		ICEPacket(ICEPacket& other);
		~ICEPacket();

		ICEPacket		&operator=(ICEPacket &other);
        
        /**
		 * Return debuggig string.
		 */
		QString			toString();

		/**
		 ** Encode the current packet data into a serialized (streamed) QByteArray
		 ** using the specified cipher. If the cipher is NULL, then
		 ** do not use a cipher.
		 */
		QByteArray&		encode(ICECrypt* crypt=NULL);
		/**
		 ** Decode the QIODevice serialized data stream into the packet data
		 ** members using the specified cipher. If the cipher is NULL, then
		 ** do not use a cipher.
		 */
		void			decode(QIODevice* io, ICECrypt* crypt=NULL);

		inline	void    			setType(PacketType type)				{mType = type;}
		inline	void    			setCode(PacketCode code)				{mCode = code;}
		inline	void				setObjectId(QByteArray id)				{mObjectId = id;}
		inline	void				setObjectType(quint32 type)				{mObjectType = type;}
		inline	void				setLogin(QString login)					{mLogin = login;}
		inline	void    			setKey(QString key)						{mKey = key;}
		inline	void    			setCipher(PacketCipher cypher)			{mCipher = cypher;}
		inline	void    			setProtocolVersion(float version)		{mProtocolVersion=version;}
		inline  void    			setPlatform(PlatformType platform)		{mPlatform=platform;}
		inline	void    			setString(QString str)					{mString=str;}
		inline	void				setInteger(qint32 n)					{mInteger=n;}
		inline	void    			setLocation(QString location)			{mLocation=location;}
		inline	void				setData(QByteArray data)				{mData=data;}
		inline	void				setUi(QByteArray ui)					{mUi=ui;}
		inline	void				setUiObjects(QStringList uiObjects)		{mUiObjects=uiObjects;}
		inline	void				setEvent(ICEEvent event)				{mEvent=event;}
        
		inline	PacketType			getType()								{return mType;}
		inline	PacketCode			getCode()								{return mCode;}
		inline	QByteArray			getObjectId()							{return mObjectId;}
		inline	quint32				getObjectType()							{return mObjectType;}
		inline	QString				getLogin()								{return mLogin;}
		inline	QString				getKey()								{return mKey;}
		inline	PacketCipher		getCipher()								{return mCipher;}
		inline	float				getProtocolVersion()					{return mProtocolVersion;}
		inline  PlatformType        getPlatform()							{return mPlatform;}
		inline	QString				getString()								{return mString;}
		inline	qint32				getInteger()							{return mInteger;}
		inline	QString				getLocation()							{return mLocation;}
		inline	QByteArray			getData()								{return mData;}
		inline	QByteArray			getUi()									{return mUi;}
		inline	QStringList			getUiObjects()							{return mUiObjects;}
        inline	ICEEvent			getEvent()								{return mEvent;}

	signals:
		void                packetError(PacketCode,QString aux);

	private:
		void                copy(ICEPacket& other);
		QDataStream         mStream;
		QByteArray          mCipherText;
		QByteArray          mClearText;
		QByteArray			mEmptyText;
		PacketType          mType;
		PacketCode          mCode;
		QByteArray			mObjectId;
		quint32				mObjectType;
		QString             mLogin;
		QString             mKey;
		PacketCipher        mCipher;
		float               mProtocolVersion;
		PlatformType        mPlatform;
		QString             mString;
		qint32				mInteger;
		QString				mLocation;
		QByteArray			mData;
		QByteArray			mUi;
		QStringList			mUiObjects;
		ICEEvent			mEvent;
};

#endif
