/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "icesession.h"
#include "icepacket.h"

#include <QByteArray>
#include <QDataStream>

QMutex	ICESession::mCountMutex;
QList<ICESession*>	ICESession::mSessions;

#define inherited QThread

ICESession::ICESession(QObject *parent, bool server)
: QThread(parent)
, mCryptTX(NULL)
, mCryptRX(NULL)
, mSocket(NULL)
, mCipher(ICEPacket::CipherXOR)
, mEncrypt(false)
, mState(SessionClosed)
, mServer(server)
, mReceiving(false)
, mSessionTimerId(-1)
, mDebug(false)
{
	qRegisterMetaType<ICESession::MessageReason>( "ICESession::MessageReason" );
	/**
	** Count session instance.
	*/
	mCountMutex.lock();
	mSessions.append( this );
	mCountMutex.unlock();
	/**
	** Start session timer.
	*/
	mSessionTimerId = startTimer(1000);
}


ICESession::~ICESession()
{
	if ( mCryptTX != NULL ) {
		delete mCryptTX;
		mCryptTX = NULL;
	}
	if ( mCryptRX != NULL ) {
		delete mCryptRX;
		mCryptRX = NULL;
	}
	mCountMutex.lock();
	mSessions.removeAt( mSessions.indexOf( this ) );
	mCountMutex.unlock();
	ICESessionObject::purge(this);
}

void ICESession::initialize()
{
	qRegisterMetaType<QAbstractSocket::SocketError>( "QAbstractSocket::SocketError" );
	qRegisterMetaType<QAbstractSocket::SocketState>( "QAbstractSocket::SocketState" );
	QObject::connect(socket(),SIGNAL(connected()),this,SLOT(slotConnected()),Qt::QueuedConnection);
	QObject::connect(socket(),SIGNAL(disconnected()),this,SLOT(slotDisconnected()),Qt::QueuedConnection);
	QObject::connect(socket(),SIGNAL(error(QAbstractSocket::SocketError)),this,SLOT(slotError(QAbstractSocket::SocketError)),Qt::QueuedConnection);
	QObject::connect(socket(),SIGNAL(hostFound()),this,SLOT(slotHostFound()),Qt::QueuedConnection);
	QObject::connect(socket(),SIGNAL(stateChanged(QAbstractSocket::SocketState)),this,SLOT(slotStateChanged(QAbstractSocket::SocketState)),Qt::QueuedConnection);
	QObject::connect(socket(),SIGNAL(readyRead()),this,SLOT(rx()),Qt::QueuedConnection);
}

void ICESession::run()
{
	if ( mServer ) {
		exec();
	}
}

void dump( QByteArray* bytes )
{
	for( int y=0; y < bytes->size(); y+= 16 ) {
		fprintf( stderr, "%04X: ", y );
		for( int x=0; x < 16 && (y+x) < bytes->size() ; x++ ) {
			fprintf( stderr, "%02X ", (unsigned char)bytes->at(y+x) );
		}
		fprintf( stderr, "\n" );
	}
}

bool ICESession::packetAvailable()
{
	if ( socket()->bytesAvailable() >= 4 ) {
		unsigned char header[4];
		if ( socket()->peek((char*)header,4) == 4 ) {
			quint32 iheader=0;
			iheader |= ((header[0] << 24) & 0xFF000000);
			iheader |= ((header[1] << 16) & 0x00FF0000);
			iheader |= ((header[2] << 8)  & 0x0000FF00);
			iheader |= ((header[3])       & 0x000000FF);
			if ( socket()->bytesAvailable() >= iheader ) {
				return true;
			}
		}
	}
	return false;
}

void ICESession::rx()
{
	if ( !mReceiving ) {
		mReceiving=true;
		while ( packetAvailable() ) {
			ICEPacket packet;
			if ( mEncrypt ) {
				packet.setKey( getKey() );
				packet.decode(socket(),mCryptRX);
			} else {
				packet.decode(socket());
			}
			if ( debugging() ) {
				emit statusMessage( this, ReasonDebugRX, packet.toString() );
			}

			rx ( packet );
		}
		mReceiving=false;
	}
}

void ICESession::tx(ICEPacket& packet)
{
	QByteArray bytes;
	if ( debugging() ) {
		emit statusMessage( this, ReasonDebugTX, packet.toString() );
	}
	if ( mEncrypt ) {
		packet.setKey( getKey() );
		bytes = packet.encode(mCryptTX);
	} else {
		bytes = packet.encode();
	}
	/* a packet has been encoded, transmit it */
	qint64 rc;
	while ( (rc = socket()->write(bytes)) != bytes.size() ) {
		if ( rc < 0 ) {
			perror("socket write");
			break;
		}
		bytes.remove(0,rc); /* re3move what we've sent already */
		socket()->flush();
	}
	socket()->flush();
}

void ICESession::cryptError(int err)
{
	/* FIXME - handle error */
	fprintf( stderr, "crypt error: " );
	switch( err ) {
	}
}

void ICESession::setEncrypt(bool encrypt)
{
	mEncrypt = encrypt;
}

void ICESession::setState(SessionState state)
{
	emit stateChanged((mState=state));
}

void ICESession::close()
{
	socket()->close();
}

void ICESession::objectGenerateNetworkHashFailed(ICESessionObject* so)
{
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| network hash function failed" ) );
}


/***************************************************************************************************
**                          Session Object Construct Request/Reply                                **
***************************************************************************************************/

void ICESession::requestSessionObjectConstruct(ICESessionObject* so)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		txpacket.setObjectType( so->type() );
		txpacket.setObjectId( so->id() );
		txpacket.setUi( so->ui() );
		txpacket.setLocation( so->location() );
		txpacket.setType ( ICEPacket::RequestSessionObjectConstruct );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object construct request" ) );
}

void ICESession::replySessionObjectConstruct( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so=NULL;
	bool obtained=false;
	QByteArray id = rxpacket->getObjectId();
	/** 
	** We did not create it, however, the object may have been constructed by a derived class.
	** So let's see if we can find it by it's object id...
	*/
	so = ICESessionObject::obtain(rxpacket->getObjectId());
	if ( so != NULL ) {
		obtained=true;
	}
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object construct request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
				/**
				** The other end may have send ui data, so let's hold onto it.
				*/
				QByteArray	ui			= rxpacket->getUi();
				QString		location	= rxpacket->getLocation();
				
				txpacket.setObjectId( so->id() );
				
				txpacket.setUi( so->ui() );
				txpacket.setLocation( so->location() );
				
				txpacket.setType ( ICEPacket::ReplySessionObjectConstruct );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
				
				/**
				** Now let's see about that ui data.
				*/
				so->setUi(ui );
				so->setLocation( location );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object construct reply" ) );
		}
		if ( obtained ) {
			so->release();
		}
	}
}

void ICESession::processReplySessionObjectConstruct( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			so->setCompleted(true);
			so->setUi( rxpacket->getUi() );
			so->setLocation( rxpacket->getLocation() );
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object construct reply" ) );
		}
		so->release(false);
	} 	
}

/***************************************************************************************************
**                          Session Object Destroy Request/Reply                                  **
***************************************************************************************************/

void ICESession::requestSessionObjectDestroy(ICESessionObject* so)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		
		txpacket.setObjectId( so->id() );
		txpacket.setType ( ICEPacket::RequestSessionObjectDestroy );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object destroy request" ) );
}

void ICESession::replySessionObjectDestroy( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId()); /** obtain a reference */
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object destroy request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
		
				txpacket.setObjectId( so->id() );
				txpacket.setType ( ICEPacket::ReplySessionObjectDestroy );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object destroy reply" ) );
			so->destroy(); /** queue the sess obj for destruction */
		}
		so->release(); /** release the reference to the sess obj */
	} 
}

void ICESession::processReplySessionObjectDestroy( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object destroy reply" ) );
		}
		so->release(false);
	} 	
}

/***************************************************************************************************
**                          Session Object Increment Reference Request/Reply                      **
***************************************************************************************************/

void ICESession::requestSessionObjectIncrementReferences(ICESessionObject* so)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		
		txpacket.setObjectId( so->id() );
		txpacket.setType ( ICEPacket::RequestSessionObjectIncrementReferences );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object obtain request" ) );
}

void ICESession::replySessionObjectIncrementReferences( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false); /** obtain a reference */
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object obtain request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
		
				txpacket.setObjectId( so->id() );
				txpacket.setType ( ICEPacket::ReplySessionObjectIncrementReferences );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object obtain reply" ) );
		}
	} 
}

void ICESession::processReplySessionObjectIncrementReferences( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object obtain reply" ) );
		}
		so->release(false);
	} 	
}

/***************************************************************************************************
**                          Session Object Decrement Reference Request/Reply                      **
***************************************************************************************************/

void ICESession::requestSessionObjectDecrementReferences(ICESessionObject* so)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		
		txpacket.setObjectId( so->id() );
		txpacket.setType ( ICEPacket::RequestSessionObjectDecrementReferences );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object release request" ) );
}

void ICESession::replySessionObjectDecrementReferences( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so = ICESessionObject::release(rxpacket->getObjectId(),false); /** release a reference */
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object release request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
		
				txpacket.setObjectId( so->id() );
				txpacket.setType ( ICEPacket::ReplySessionObjectDecrementReferences );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object release reply" ) );
		}
	} 
}

void ICESession::processReplySessionObjectDecrementReferences( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object release reply" ) );
		}
		so->release(false);
	} 	
}

/***************************************************************************************************
**                          Session Object Ui Data Request/Reply                                **
***************************************************************************************************/
 
void ICESession::requestSessionObjectUiObjects(ICESessionObject* so)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		
		txpacket.setObjectId( so->id() );
		txpacket.setUiObjects( so->uiObjects() );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object ui objects request" ) );
}

void ICESession::replySessionObjectUiObjects( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object ui objects request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
				
				QStringList uiObjects = rxpacket->getUiObjects(); /** requester may have sent some data */		
				
				txpacket.setObjectId( so->id() );
				txpacket.setUiObjects( so->uiObjects() );
				txpacket.setType ( ICEPacket::ReplySessionObjectUiObjects );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
				
				so->setUiObjects( uiObjects );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object ui objects reply" ) );
		}
		so->release(false);
	} 
}

void ICESession::processReplySessionObjectUiObjects( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			if ( state() == ICESession::SessionValidated )
			{
				so->setUiObjects( rxpacket->getUiObjects() );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object ui objects reply" ) );
		}
		so->release(false);
	} 	
}

/***************************************************************************************************
**                          Session Object Ui Request/Reply                                     **
***************************************************************************************************/
 
void ICESession::requestSessionObjectUi(ICESessionObject* so)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		
		txpacket.setObjectId( so->id() );
		txpacket.setUi( so->ui() );
		txpacket.setType ( ICEPacket::RequestSessionObjectUi );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object ui request" ) );
}

void ICESession::replySessionObjectUi( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object ui request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
				
				QByteArray ui = rxpacket->getUi(); /** requester may have sent some data */		
				
				txpacket.setObjectId( so->id() );
				txpacket.setUi( so->ui() );
				txpacket.setType ( ICEPacket::ReplySessionObjectUi );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
				
				so->setUi( ui );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object ui reply" ) );
		}
		so->release(false);
	} 
}

void ICESession::processReplySessionObjectUi( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			if ( state() == ICESession::SessionValidated )
			{
				so->setUi( rxpacket->getUi() );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object ui reply" ) );
		}
		so->release(false);
	} 	
}

/***************************************************************************************************
**                          Session Object Event Request/Reply                                   **
***************************************************************************************************/
 
void ICESession::requestSessionObjectEvent(ICESessionObject* so, ICEEvent event)
{
	if ( state() == ICESession::SessionValidated )
	{
		ICEPacket txpacket;
		
		txpacket.setObjectId( so->id() );
		txpacket.setEvent( event );
		txpacket.setType ( ICEPacket::RequestSessionObjectEvent );
		txpacket.setCode ( ICEPacket::CodeOK );
		tx ( txpacket );
	}
	emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object event request" ) );
}

void ICESession::replySessionObjectEvent( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object event request" ) );
			if ( state() == ICESession::SessionValidated )
			{
				ICEPacket txpacket;
				txpacket.setObjectId( so->id() );
				txpacket.setType ( ICEPacket::ReplySessionObjectEvent );
				txpacket.setCode ( ICEPacket::CodeOK );
				tx ( txpacket );
			}
			emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| >> session object event reply" ) );
		}
		so->release(false);
	} 
}

void ICESession::processReplySessionObjectEvent( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			if ( state() == ICESession::SessionValidated )
			{
				emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object event reply" ) );
		
			}
		}
		so->release(false);
	} 	
}

/**
	\brief Dispatch incoming packet to appropriate handler.
	\param rxpacket Inbound packet.
	\return void
 **/
void ICESession::rx ( ICEPacket rxpacket )
{
	switch ( rxpacket.getType() )
	{
		case ICEPacket::RequestSessionObjectConstruct:				replySessionObjectConstruct ( &rxpacket ); break;
		case ICEPacket::RequestSessionObjectDestroy:				replySessionObjectDestroy ( &rxpacket );   break;
		case ICEPacket::RequestSessionObjectIncrementReferences:	replySessionObjectIncrementReferences ( &rxpacket ); break;
		case ICEPacket::RequestSessionObjectDecrementReferences:	replySessionObjectDecrementReferences ( &rxpacket ); break;
		case ICEPacket::RequestSessionObjectUi:						replySessionObjectUi( &rxpacket ); break;
		case ICEPacket::RequestSessionObjectUiObjects:				replySessionObjectUiObjects( &rxpacket ); break;
		case ICEPacket::RequestSessionObjectEvent:					replySessionObjectEvent( &rxpacket ); break;
		
		case ICEPacket::ReplySessionObjectConstruct:				processReplySessionObjectConstruct ( &rxpacket ); break;
		case ICEPacket::ReplySessionObjectDestroy:					processReplySessionObjectDestroy ( &rxpacket );   break;
		case ICEPacket::ReplySessionObjectIncrementReferences:		processReplySessionObjectIncrementReferences ( &rxpacket ); break;
		case ICEPacket::ReplySessionObjectDecrementReferences:		processReplySessionObjectDecrementReferences ( &rxpacket ); break;
		case ICEPacket::ReplySessionObjectUi:						processReplySessionObjectUi( &rxpacket ); break;
		case ICEPacket::ReplySessionObjectUiObjects:				processReplySessionObjectUiObjects( &rxpacket ); break;
		case ICEPacket::ReplySessionObjectEvent:					processReplySessionObjectEvent( &rxpacket ); break;
		default:
			emit statusMessage ( this,ICESession::ReasonSession,tr ( "Bad Packet Type " ) +QString::number ( rxpacket.getType() ) );
			//socket()->close();
			break;
	}
}

void ICESession::timerEvent(QTimerEvent* e)
{
	inherited::timerEvent(e);
	if (e->timerId() == mSessionTimerId) {
		if ( socket() != NULL ) {
			socket()->flush();
			rx();
		}
	}
}
