/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ICESESSION_H
#define ICESESSION_H

#include <QObject>
#include <QTcpSocket>
#include <QThread>
#include <QTimerEvent>
#include <QUrl>
#include <QTimer>
#include <QMutex>

#include "icepacket.h"

#include "icecrypt.h"
#include "icecryptbf.h"
#include "icecryptseal.h"
#include "icecryptloki91.h"
#include "iceevent.h"

#include "icesessionobject.h"

/**
	@author Michael Sharkey
*/
class ICESession : public QThread
{
	Q_OBJECT
	public:
		typedef enum {
			ReasonUnknown=0,
			ReasonVersion,
			ReasonServer,
			ReasonSession,
   			ReasonSessionObj,
			ReasonDirectory,
			ReasonDatabase,
			ReasonDebug,
			ReasonDebugTX,
			ReasonDebugRX
		} MessageReason;
    
		typedef enum {
			SessionClosed=0,
			SessionConnected=1,
			SessionChallenge,
			SessionValidated,
			SessionClosing
		} SessionState;

		typedef enum {
			OperationOK=0,
			OperationNotPermitted=1000,
			OperationNotSupported,
		} OperationCode;

		ICESession(QObject *parent = 0, bool server=true );
		virtual ~ICESession();
		void setDebugging(bool db)  {mDebug=db;}
		bool debugging()    {return mDebug;}

		QTcpSocket*     socket() {return mSocket;}
		SessionState    state() {return mState;}

		void        setLogin(QString login) {mLogin = login;}
		void        setKey(QString key) {mKey = key;}

		QString     getLogin() {return mLogin;}
		QString     getKey() {return mKey;}

		void   setCipher(ICEPacket::PacketCipher cipher) {mCipher=cipher;}
		ICEPacket::PacketCipher getCipher() {return mCipher;};

		virtual     void        rx(ICEPacket packet);
		
		static	quint32 count() {return mSessions.count();}
		static QList<ICESession*>* sessions() {return &mSessions;}	

		
		virtual		ICESessionObject* construct(ICESessionObject::ObjectType,QByteArray* id=NULL,QByteArray* data=NULL)=0;
	
	public slots:
		virtual void	run();
		virtual void    asyncPopupMessage(QString text,ICEPacket::PacketCode code=ICEPacket::CodeOK)=0;
		virtual	void	requestSessionObjectConstruct(ICESessionObject* so);
		virtual	void	requestSessionObjectDestroy(ICESessionObject* so);
		virtual	void	requestSessionObjectIncrementReferences(ICESessionObject* so);
		virtual	void	requestSessionObjectDecrementReferences(ICESessionObject* so);
		virtual	void	requestSessionObjectUi(ICESessionObject* so);
		virtual	void	requestSessionObjectUiObjects(ICESessionObject* so);
		virtual	void	requestSessionObjectEvent(ICESessionObject* so,ICEEvent event);
		
		virtual	void	objectGenerateNetworkHashFailed(ICESessionObject* so);
				
				void	sessionObjectStatusMessage(ICESessionObject* so,QString message) {
					emit statusMessage ( this, ICESession::ReasonSessionObj,so->toString() +  message );
				}
	
	signals:
		void		statusMessage(QString message);
		void		statusMessage(void*,ICESession::MessageReason reason,QString message);
		void        operationError(ICESession::OperationCode);
		void        stateChanged(ICESession::SessionState);
		void        packetError(ICEPacket::PacketCode,QString aux);

	protected:
		
		virtual		void		replySessionObjectConstruct(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		virtual		void		replySessionObjectDestroy(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		virtual		void		replySessionObjectIncrementReferences(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		virtual		void		replySessionObjectDecrementReferences(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		virtual		void		replySessionObjectUi(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		virtual		void		replySessionObjectUiObjects(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		virtual		void		replySessionObjectEvent(ICEPacket*,ICEPacket::PacketCode code=ICEPacket::CodeOK);
		
		virtual		void		processReplySessionObjectConstruct(ICEPacket*);
		virtual		void		processReplySessionObjectDestroy(ICEPacket*);
		virtual		void		processReplySessionObjectIncrementReferences(ICEPacket*);
		virtual		void		processReplySessionObjectDecrementReferences(ICEPacket*);
		virtual		void		processReplySessionObjectUi(ICEPacket*);
		virtual		void		processReplySessionObjectUiObjects(ICEPacket*);
		virtual		void		processReplySessionObjectEvent( ICEPacket* rxpacket );
		
					bool		packetAvailable();
					void		initialize();
		virtual     void		tx(ICEPacket& packet);
		virtual     void		setState(SessionState state);
		virtual     void		setEncrypt(bool encrypt);
		virtual		void		timerEvent(QTimerEvent* e);
		ICECrypt*				mCryptTX;
		ICECrypt*				mCryptRX;
		QTcpSocket*				mSocket;
		ICEPacket::PacketCipher mCipher;
		int						mSessionTimerId;

	protected slots:
		virtual     void        close();
		virtual     void        rx();                           /* receive pending data */
		virtual     void        cryptError(int);
		virtual     void        slotConnected() {}
		virtual     void        slotDisconnected()=0;
		virtual     void        slotError(QAbstractSocket::SocketError err)=0;
		virtual     void        slotHostFound() {}
		virtual     void        slotStateChanged(QAbstractSocket::SocketState state)=0;

	private:
		static	QMutex				mCountMutex;
		static	QList<ICESession*>	mSessions;
		bool                    	mEncrypt;
		QString                 	mLogin;
		QString                 	mKey;
		SessionState				mState;
		bool						mServer;
		bool						mReceiving;
		bool                    	mDebug;
};

#endif
