/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "icesession.h"
#include "icesessionobject.h"
#include "icemd5.h"

#include <QNetworkInterface>
#include <QHostAddress>
#include <QDateTime>
#include <QString>
#include <QFileInfo>

#include "assert.h"

quint64	ICESessionObject::mIDSequence=0;

QMutex	ICESessionObject::mDestroyedListMutex;
QMutex	ICESessionObject::mHashMutex;
QMutex	ICESessionObject::mIDMutex;
QString	ICESessionObject::mNetworkHash;

QHash<QByteArray,ICESessionObject*>		ICESessionObject::mHash;
QList<ICESessionObject*>				ICESessionObject::mDestroyedList;

#define inherited QObject

ICESessionObject::ICESessionObject(ICESession *session,QByteArray* id)
 : inherited(session)
 , mReferenceCount(0)
 , mCompleted(false)
{
	assert(session!=NULL);
	if ( id==NULL ) {
		generateID();
		mHashMutex.lock();
		mHash.insert(ICESessionObject::id(),this);
	} else {
		mHashMutex.lock();
		mHash.insert((mID=*id),this);
	}
	mHashMutex.unlock();
	QObject::connect ( this,SIGNAL ( statusMessage ( ICESessionObject*,QString ) ),session,SLOT ( sessionObjectStatusMessage ( ICESessionObject*,QString ) ) );
	QObject::connect(this,SIGNAL(objectDestroyed(ICESessionObject*)),session,SLOT(requestSessionObjectDestroy(ICESessionObject*)));
	
}


ICESessionObject::~ICESessionObject()
{
	/**
	** We should take this out from the list prior to emitting so that
	** there is no recursive deletes.
	*/
	
	for(int i=0; i < mScripts.count(); i++ )
	{
		ICEScriptThread* script = mScripts[i];
		script->quit();
		if ( !script->wait(1000) )
		{
			script->terminate();
		}
		delete script;
	}
	
	mHashMutex.lock();
	mHash.remove(id());
	mHashMutex.unlock();
	
	mDestroyedListMutex.lock();
	mDestroyedList.removeAll(this);
	mDestroyedListMutex.unlock();
	
	emit objectDestroyed(this);
}

void ICESessionObject::alert( QString msg )
{
	//emit statusMessage( this, "alert: \""+msg+"\"" );
	session()->asyncPopupMessage( msg );
}


void ICESessionObject::appendChild(QByteArray childId)
{
	if ( mChildrenIds.indexOf(childId) < 0 ) {
		mChildrenIds.append(childId);
	}
}

void ICESessionObject::removeChild(QByteArray childId)
{
	mChildrenIds.removeAll(childId);
}

void ICESessionObject::removeParent()
{
	mParentID = QByteArray();
}

QList<QByteArray> ICESessionObject::allChildren()
{
	QList<QByteArray> all=mChildrenIds;
	for( int i = 0; i < mChildrenIds.count(); i++ ) {
		ICESessionObject* child = obtain( mChildrenIds[i] );
		if ( child != NULL ) {
			all += child->allChildren();
			child->release();
		}
	}
	return all;
}

void ICESessionObject::setParent(QByteArray parentId)
{
	if ( !parentId.isEmpty() ) {
		ICESessionObject* soParent = obtain( parentId );
		if ( soParent != NULL ) {
			soParent->appendChild(id());
			mParentID = parentId;
			soParent->release();
		}
	}
}

/**
** \brief Create a unique session object ID. Only the original (master) constructor calls this.
** FIXME Add system time and process id to the hash string.
*/
void ICESessionObject::generateID()
{
	mIDMutex.lock();	/** No multiple sessions manipulating this data at the same time. */
	
	/**
	** If we have not already cache'd a network hash string, then let's do so now...
	*/
	if ( mNetworkHash.isEmpty() ) {
		QList<QNetworkInterface> interfaces = QNetworkInterface::allInterfaces();
		QList<QHostAddress> hostAddresses = QNetworkInterface::allAddresses();
		/**
		** First, concat the interface names and MACs...
		*/
		for( int i = 0; i < interfaces.count(); i++ ) {
			QNetworkInterface interface = interfaces.at(i);
			if ( interface.isValid() ) {
				mNetworkHash += interface.name() + interface.hardwareAddress();
			}
		}
		/**
		** Then, concat the interface addresses...
		*/
		for( int i = 0; i < hostAddresses.count(); i++ ) {
			QHostAddress hostAddress = hostAddresses.at(i);
			if ( !hostAddress.isNull() ) {
				mNetworkHash += hostAddress.toString();
			}
		}
	}
	
	/**
	** Generate the ID, ensuring it is unique...
	*/
	QByteArray newID;
	//do {
		ICEMD5 md5( 
				mNetworkHash + 
				QString::number((quint64)this) + 
				QString::number(++mIDSequence==0?++mIDSequence:mIDSequence) +
				QDateTime::currentDateTime().toString()
				);
		unsigned char* raw_digest = md5.raw_digest();
		newID = QByteArray( (char*)raw_digest, 16 );
		delete raw_digest;
	//} while ( find( newID ) != NULL );
	mID = newID;
	
	mIDMutex.unlock();	/** No multiple sessions manipulating this data at the same time. */

}

/**
** \brief Returns the count of session objects.
*/
int ICESessionObject::count(ICESession* session) 
{
	if ( session != NULL )  {
		int cnt=0;
		QHash<QByteArray,ICESessionObject*>::const_iterator i = mHash.begin();
		while (i != mHash.end()) {
			if (i.value()->session() == session) {
				++cnt;
			}
			++i;
		}
		return cnt;	
	}
	return mHash.keys().count();
}

/**
** \brief Returns the session object id as an ascii hex string.
*/
QString ICESessionObject::hexID()
{
	QString rc;
	for( int i = 0; i < id().count(); i++ ) {
		unsigned char ch = id().at(i);
		char hex[3];
		sprintf( hex,"%02X", ch );
		rc += QString( hex );
	}
	return rc;
}

/**
** \brief Report session object data as a string
*/
QString ICESessionObject::toString()
{
	char hexbuf[20];
	char decbuf[20];	
	/** Format the session object memory address... */
	switch(sizeof(char*)) {
		case 3:
			case 4:		sprintf( hexbuf, "%08p",  (char*)this ); break;
			case 8:		sprintf( hexbuf, "%016p", (char*)this ); break;
			case 16:	sprintf( hexbuf, "%032p", (char*)this ); break;
			default:	sprintf( hexbuf, "%p",     (char*)this ); break;
	}
	sprintf( decbuf, "%04d", referenceCount() );
	QString rc( decbuf );
	sprintf( decbuf, "%04d", count(session()) );
	QString ct( decbuf );
	QString str = hexID() + "|"  + 
			QString( hexbuf ) + 
			"|rc(" + rc + ")" +
			"|ct(" + ct + ")";
	return str;
}

/**
** \brief Most likely the session is being deleted so purge all objects owned by this session.
*/
void ICESessionObject::purge(ICESession* session)
{
	while ( mDestroyedList.count() > 0 ) {
		ICESessionObject* so = mDestroyedList.first();
		if ( so != NULL ) {
			if ( session == NULL || (session!=NULL && so->session()==session) ) {
				delete so;
			}
		}
	}
}

/**
** \brief Increment reference count.
*/
void ICESessionObject::incrementReferences(bool remote)
{
	mReferenceMutex.lock();
	++mReferenceCount;
	mReferenceMutex.unlock();
	if ( remote ) {
		session()->requestSessionObjectIncrementReferences(this);
	}
}

/**
** \brief Decrement reference count.
*/
void ICESessionObject::decrementReferences(bool remote)
{
	mReferenceMutex.lock();
	if ( --mReferenceCount < 0 ) {
		emit statusMessage ( this,QString( "| reference count underflow" ) );
	}
	mReferenceMutex.unlock();
	if ( remote ) {
		session()->requestSessionObjectDecrementReferences(this);
	}
}

/**
** \brief Find a pointer to a session object using the session object byte array.
*/
ICESessionObject* ICESessionObject::find(QByteArray id) 
{
	QHash<QByteArray,ICESessionObject*>::const_iterator i = mHash.find(id);
	if (i != mHash.end() && i.key() == id) {
		return i.value();
	}
	return NULL;
}

/**
** \brief Obtain a reference to a session object. This increments the reference count.
*/
ICESessionObject* ICESessionObject::obtain(QByteArray id,bool remote)
{
	ICESessionObject* so = find(id);
	if ( so != NULL ) {
		return so->obtain(remote);
	}
	return NULL;
}

/**
 ** \brief Obtain a reference to a session object. This increments the reference count.
 */
ICESessionObject* ICESessionObject::obtain(bool remote)
{
	if ( !destroyed() ) {
		incrementReferences(remote);
		return this;
	}
	return NULL;
}

/**
** \brief Release a reference to a session object.
*/
ICESessionObject* ICESessionObject::release(QByteArray id,bool remote)
{
	ICESessionObject* so = find(id);
	if ( so != NULL ) {
		so->release(remote);
	}
	return so;
}

/**
 ** \brief Release a reference to a session object.
 */
ICESessionObject* ICESessionObject::release(bool remote)
{
	decrementReferences(remote);
	return this;
}

/**
 **	\brief Queue a session object for destruction. No more reference may be obtained.
 */
void ICESessionObject::destroy(QByteArray id)
{
	ICESessionObject* so = find(id);
	if ( so != NULL ) {
		so->destroy();
	}
}

/**
 **	\brief Queue a session object for destruction. No more reference may be obtained.
 */
void ICESessionObject::destroy()
{
	QList<QByteArray> childrenIds = children();
	for( int i=0; i < childrenIds.count(); i++ ) {
		QByteArray childId = childrenIds[i];
		destroy(childId);
	}
	mDestroyedListMutex.lock();
	if (!mDestroyedList.contains(this)) {
		mDestroyedList.append(this);
	}
	mDestroyedListMutex.unlock();
}

/**
 **	\brief Determine if a session object has been queued for destruction. No more reference may be obtained.
 */
bool ICESessionObject::destroyed(QByteArray id)
{
	bool rc=true;
	ICESessionObject* so = find(id);
	if ( so != NULL ) {
		return so->destroyed();
	}
	return rc;	
}

/**
 **	\brief Determine if a session object has been queued for destruction. No more reference may be obtained.
 */
bool ICESessionObject::destroyed()
{
	bool rc;
	mDestroyedListMutex.lock();
	rc = mDestroyedList.contains(this);
	if ( rc && referenceCount() > 0 ) {
		rc = false; /** if we're being destroyed but we still have references, then we are not destroyed yet. */
	}
	mDestroyedListMutex.unlock();
	return rc;
}

/**
 **	\brief Get a reference count.
 */
int ICESessionObject::referenceCount(QByteArray id)
{
	ICESessionObject* so = find(id);
	if ( so != NULL ) {
		return so->referenceCount();
	}
	return 0;	
}


