/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ICESESSIONOBJECT_H
#define ICESESSIONOBJECT_H

#include <QObject>
#include <QDataStream>
#include <QByteArray>
#include <QHash>
#include <QList>
#include <QMutex>
#include <QScriptable>
#include <QStringList>

#include "icescriptthread.h"

/**
	Base class for objects shared between server and client.

	@author R. Michael Sharkey <mike@pikeaero.com>
*/

class ICESession;
class ICESessionObject : public QObject, protected QScriptable
{
	Q_OBJECT
	public:
		typedef enum {
			/**
			** System level session object types...
			*/
			TypeInvalid=0,				/** Invalid object type */

			/**
			** User interface(Ui) transport session object types...
			*/
			TypeUiTransport=1961,		/** An object to contain the Ui */
			TypeUiElementTransport,
   
		} ObjectType;
		
		/**
		** Constructors, destructors, and copy operators...
		*/
		ICESessionObject(ICESession *session,QByteArray *id=NULL);
		virtual ~ICESessionObject();
		
		/**
		** Constructor has completed? (both ends) 
		*/
		inline	bool completed() {return mCompleted;}
		inline	void setCompleted(bool c) {mCompleted=c;}
		
		/** 
		** Get object id. 
		*/
		inline	const QByteArray&	id() const						{ return mID; }
		inline	const QByteArray&	parentId() const				{ return mParentID; }
		void	setParent(QByteArray parentId);
		void	appendChild(QByteArray childId);
		void	removeChild(QByteArray childId);
		void	removeChildren();
		void	removeParent();
		QList<QByteArray> children() 								{ return mChildrenIds; }
		QList<QByteArray> allChildren();
		QString	hexID();
		
		/**
		 ** Common client/server access methods....
		 */
		virtual ObjectType type()=0;
		virtual	QByteArray		ui()							{return QByteArray();}
		virtual QStringList		uiObjects()						{return QStringList();}
		virtual	QString			location()						{return QString();}
		virtual	ICEScriptThread* setLocation(QString location)	{}
		
		/** 
		 ** Report session object data as a string. 
		 */
		virtual QString	toString();
		
		/** 
		** Set object id...CAUTION! Do not use unless you know what you are doing! 
		*/
		inline	QByteArray&	setID(QByteArray&  id)		{ return (mID=id); }
		
		/** 
		** Return a pointer to the session. 
		*/
		inline	ICESession*	session() const				{ return (ICESession*)parent(); }
		
		/** 
		** Return a count session objects 
		*/
		static int count(ICESession* session=NULL);
			
		/**
		** Obtain a reference to a session object. This guarantees that the session
		** object will not be destroyed while there are outstanding references.
		** A session object reference cannot be obtained if it is pending to be
		** destroyed, in this case NULL is returned. Always refer to the
		** returned pointer after calling "obtain".
		*/
				ICESessionObject* obtain(bool remote=true);	
		static	ICESessionObject* obtain(QByteArray id,bool remote=true);
		/**
		 ** Release a refernce to a session object. This guarantees that the session
		 ** object will not be destroyed while there are outstanding references.
		 */
				ICESessionObject* release(bool remote=true);	
		static	ICESessionObject* release(QByteArray id,bool remote=true);
		/**
		** Queue a session object for destruction. The object will not be destroyed
		** until the last obtained reference is released.
		*/
				void destroy();
		static 	void destroy(QByteArray id);
		/**
		 ** Determine if a session object is queued to be destroyed.
		 */
				bool destroyed();
		static	bool destroyed(QByteArray id);
		/**
		 ** Return a reference count for this object.
		 */
		inline	int referenceCount() {return mReferenceCount;}
		static	int	referenceCount(QByteArray id);
		
				
		/**
		 ** Purge all objects which are pending destruction.
		 */
		static void purge(ICESession* session=0);
		
		
	signals:
		void objectDestroyed(ICESessionObject*);
		void statusMessage(ICESessionObject*, QString message);
		
	public slots:
		virtual	void	alert( QString msg ); /** send an alert across the session object */
		virtual void	setUi(QByteArray ui) {}
		virtual void	setUiObjects(QStringList uiObjects) {}
				void	incrementReferences(bool remote=true);
				void	decrementReferences(bool remote=true);
	
	protected:
		/**
		 ** Script thread list...
		 */
		QList<ICEScriptThread*>	mScripts;
				
	private:
		/** 
		** Find a session object by id and return a pointer to it, or NULL. 
		*/
		static ICESessionObject* find(QByteArray id);
		/**
		** List of session objects queued for destruction...
		*/
		static QMutex	mDestroyedListMutex;
		static QList<ICESessionObject*>	mDestroyedList;
		/**
		** Reference count and reference count mutex.
		*/
		QMutex 	mReferenceMutex;	/** Locks refernce coubnt */
		int		mReferenceCount;	/** number of references */
		/**
		** List of session objects. Using a QHash table for faster lookups than
		** when using a QList....FIXME performance of looking up a session object by it's
		** id is critical - do some tesing here - may require implementing
		** a b-tree to obtain very fast lookups.
		*/
		static QMutex 	mHashMutex;							/** Locks session object table */
		static QHash<QByteArray,ICESessionObject*>	mHash;	/** The session objects */
		/**
		** Object id's must be unique between both ends of the connection, therefore they are
		** MD5 sequences generated from a hash string using identifiers from the network interface
		** such as MAC addresses and IP addresses, with a sequence number appended.
		**/
		void generateID();					/** Generate mID for this instance. */
		QByteArray			mID;			/** 16 byte MD5 array as unique object identifier for self */
		QByteArray			mParentID;		/** 16 byte MD5 array as unique object identifier for parent sess obj */
		QList<QByteArray>	mChildrenIds;	/** list of children */
		static QMutex 		mIDMutex;		/** Locks ID generation data */
		static quint64		mIDSequence;	/** Object ID Sequence number for generating unique id */
		static QString		mNetworkHash;	/** Cache hash string for generating unique MD5 id */
		/**
		** Constructor...
		*/
		bool				mCompleted;		/** Constructor has completed? (both ends) */
		
};

#endif
