/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "iceserver.h"

#include <QCoreApplication>
#include <QHostAddress>
#include <QThread>

#include "iceserver.h"
#include "iceservertcp.h"
#include "iceserversession.h"
#include "iceserversettings.h"
#include "icesessionobject.h"

#include <stdlib.h>

extern ICEServerSettings serverSettings;

#define inherited QObject

QList<ICEServerSession*> ICEServer::mSessionList;

ICEServer::ICEServer ( QObject* parent )
: QObject ( parent )
, mInitialized ( false )
{
	mSlowTimer = startTimer ( 1000 );
	mFastTimer = startTimer ( 50 );
}

ICEServer::~ICEServer()
{
	mInitialized=false;
}

/**
** Main function is to connect to the database. Also emit some information messages
** to the log output.
*/
void ICEServer::initialize()
{
	if ( !mInitialized )
	{
		mServerTcp.setMaxPendingConnections ( serverSettings.getBacklog() );
		QObject::connect ( &mServerTcp,SIGNAL ( newConnection() ),this,SLOT ( newConnection() ) );
		QObject::connect ( this,SIGNAL ( statusMessage ( void*,ICESession::MessageReason,QString ) ),
		                   this,SLOT ( slotStatusMessage ( void*,ICESession::MessageReason,QString ) ) );
		QObject::connect ( &mDatabase,SIGNAL ( statusMessage ( void*,ICESession::MessageReason,QString ) ),
		                   this,SIGNAL ( statusMessage ( void*,ICESession::MessageReason,QString ) ) );
		QObject::connect ( &mSQL,SIGNAL ( statusMessage ( void*,ICESession::MessageReason,QString ) ),
							this,SIGNAL ( statusMessage ( void*,ICESession::MessageReason,QString ) ) );
		emit statusMessage ( this, ICESession::ReasonVersion, SERVER_NAME + tr ( " Version " ) + SERVER_VERSION );
		emit statusMessage ( this, ICESession::ReasonVersion, SERVER_COPYRIGHT );
		emit statusMessage ( this, ICESession::ReasonVersion, SERVER_RIGHTS );
		emit statusMessage ( this, ICESession::ReasonVersion, SERVER_CONTACT );
		emit statusMessage ( this, ICESession::ReasonServer,tr ( "word size is " ) + QString::number ( sizeof(char*)*8 ) + tr ( " bits" ) );
		if ( mSQL.load( serverSettings.getQueryRoot() ) ) {
			if ( mDatabase.login ( serverSettings.getSqlServer(),
									serverSettings.getSqlDatabase(), 
									serverSettings.getSqlUser(),
									serverSettings.getSqlPassword() ) )
			{
				bool rc=true;
				if ( serverSettings.getSqlInitialize() ) {
					if ( mDatabase.initialize() ) {
						serverSettings.setSqlInitialize(false);
						serverSettings.save();
					} else {
						emit statusMessage( &mDatabase, ICESession::ReasonDatabase, "Database Initialization failed!" );
						rc=false;
					}
				}
				if ( rc ) {
					mServerTcp.listen ( QHostAddress::Any, serverSettings.getPort() );
					emit statusMessage ( this, ICESession::ReasonServer,tr ( "listening on port " ) +QString::number ( serverSettings.getPort() ) );
					emit statusMessage ( this, ICESession::ReasonServer,tr ( "logging level " ) +QString::number ( serverSettings.getLogging() ) );
				}
			}
			else
			{
				emit statusMessage( &mDatabase, ICESession::ReasonDatabase, mDatabase.lastError().databaseText() );
				emit statusMessage( &mDatabase, ICESession::ReasonDatabase, mDatabase.lastError().driverText() );
			}
		} else {
			emit statusMessage( &mSQL, ICESession::ReasonDatabase, "Failed to load SQL queries from \"" + serverSettings.getQueryRoot() + "\"!" );
		}
		emit statusMessage ( this, ICESession::ReasonVersion, serverSettings.getServerName() );
		mInitialized=true;
	}
}

void ICEServer::sessionClosed()
{
	for ( int index=0; index < mSessionList.count(); index++ )
	{
		ICEServerSession* session = mSessionList.at ( index );
		if ( session->isFinished() )
		{
			mSessionList.removeAt ( index );
			emit statusMessage ( session, ICESession::ReasonSession,tr ( "closed session" ) );
			delete session;
		}
	}
}

void ICEServer::spawnNewSession ( QTcpSocket* socket )
{
	ICEServerSession* session = new ICEServerSession ( this,socket,&mDatabase );
	mSessionList.append ( session );
	QObject::connect ( session,SIGNAL ( finished() ),this,SLOT ( sessionClosed() ) );
	QObject::connect ( session,SIGNAL ( statusMessage ( void*,ICESession::MessageReason,QString ) ),
	                   this,SLOT ( slotStatusMessage ( void*,ICESession::MessageReason,QString ) ) );
	emit statusMessage ( session, ICESession::ReasonSession,tr ( "new session (# sessions=" ) +QString::number ( mSessionList.count() ) +QString ( ")" ) );
	session->start();
}

void ICEServer::newConnection()
{
	QTcpSocket* socket;
	while ( ( socket = mServerTcp.nextPendingConnection() ) != NULL )
	{
		spawnNewSession ( socket );
	}
}

void ICEServer::slotStatusMessage ( void* session, ICESession::MessageReason reason, QString msg )
{
	char buf[128];
	
	/** Format the session address... */
	switch(sizeof(char*)) {
		case 3:
		case 4:		sprintf( buf, "%08p",  (char*)session ); break;
		case 8:		sprintf( buf, "%016p", (char*)session ); break;
		case 16:	sprintf( buf, "%032p", (char*)session ); break;
		default:	sprintf( buf, "%p",    (char*)session ); break;
	}
	QString reason_str;
	QString direction_str;
	if ( (int)reason <= serverSettings.getLogging() ) {
		switch ( reason )
		{
			default:
			case ICESession::ReasonUnknown:   reason_str = "UNKNOWN  "; break;
			case ICESession::ReasonVersion:   reason_str = "VERSION  "; break;
			case ICESession::ReasonServer:    reason_str = "SERVER   "; break;
			case ICESession::ReasonSession:   reason_str = "SESSION  "; break;
			case ICESession::ReasonSessionObj:reason_str = "SESS_OBJ "; break;
			case ICESession::ReasonDirectory: reason_str = "DIRECTORY"; break;
			case ICESession::ReasonDatabase:  reason_str = "DATABASE "; break;
			case ICESession::ReasonDebug:     reason_str = "DEBUG    "; break;
			case ICESession::ReasonDebugTX:   reason_str = "DEBUG_TX "; direction_str=">> "; break;
			case ICESession::ReasonDebugRX:   reason_str = "DEBUG_RX "; direction_str="<< "; break;
		}
		QString log_msg = QDateTime::currentDateTime().toString ( Qt::ISODate ) + QString ( "|" ) +
						reason_str + QString ( "|" ) +
						QString ( buf ) + QString ( "| " ) +
						direction_str + msg;
	
		printf ( "%s\n", log_msg.toLatin1().data() );
	}
}

void ICEServer::timerEvent ( QTimerEvent* e )
{
	if ( e->timerId() == mSlowTimer )
	{
		if ( !mInitialized )
		{
			initialize();
		}
		if ( !mServerTcp.isListening() )
		{
			/**
			* Something has stopped the server from listening on the port...
			*/
			fprintf ( stderr, "%s\n",  mServerTcp.errorString().toAscii().data() );
			emit quit();
		}
	} else if ( e->timerId() == mFastTimer ) {
		/**
		 ** Periodically purge session objects...
		 */
		ICESessionObject::purge();
	}
}


void ICEServer::readSettings()
{
	serverSettings.restore();
}

void ICEServer::writeSettings()
{
	serverSettings.save();
}
