/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "iceserver.h"
#include "iceserversettings.h"
#include "icesql.h"

#include <QCoreApplication>
#include <QString>
#include <stdlib.h>

ICEServerSettings serverSettings;

static bool save_settings=true;
static QString settings_file;

static void print_version()
{
	printf( "%s Version %s\n", SERVER_NAME, SERVER_VERSION );
	printf( "%s\n", SERVER_COPYRIGHT );
	printf( "%s\n", SERVER_RIGHTS );
	printf( "%s\n", SERVER_CONTACT );
	printf( "Site name is \"%s\"\n", serverSettings.getServerName().toAscii().data() );
}

static void print_help()
{
	printf( "** Transient Settings:\n\n" );
	printf( "--version     -v       Print Version\n" );
	printf( "--nosave      -n       Don't save command line setting (default=save)\n" );
	printf( "--settings    -s <n>   Settings file to use\n" );
	printf( "--help        -h -?    Print this help\n");
	printf( "\n** Semi-persistent settings (persist until successful):\n\n" );
	printf( "--formsroot   -f <n>   Load forms from directory (current=%s)\n", serverSettings.getFormsRoot().toAscii().data() );
	printf( "--sqlinitialize        Initialize SQL tables (current=%s)\n", serverSettings.getSqlInitialize()?"true":"false" );
	printf( "\n** Persistent settings:\n\n" );
	printf( "--sqltype <n>          SQL database type (current=%s)\n", serverSettings.getSqlType().toAscii().data() );
	printf( "--sqlserver <n>        SQL server name (current=%s)\n", serverSettings.getSqlServer().toAscii().data() );
	printf( "--sqldatabase <n>      SQL dtabase name (current=%s)\n", serverSettings.getSqlDatabase().toAscii().data() );
	printf( "--sqluser <n>          SQL login id (current=%s)\n", serverSettings.getSqlUser().toAscii().data() );
	printf( "--sqlpassword <n>      SQL Database password\n" );
	printf( "--port        -p <n>   Set TCP port (current=%d)\n", serverSettings.getPort() );
	printf( "--backlog     -b <n>   Set backlog (current=%d)\n", serverSettings.getBacklog() );
	printf( "--logging     -l <n>   Set Logging Level 0..9 (current=%d)\n", serverSettings.getLogging() );
	printf( "--dataroot    -d <n>   Root folder for data storage (current=%s)\n", serverSettings.getDataRoot().toAscii().data() ); 
	printf( "--queryroot   -q <n>   Root folder for SQL query storage (current=%s)\n", serverSettings.getQueryRoot().toAscii().data() ); 
	printf( "--servername  -S <n>   Set the name of the server (current=%s)\n", serverSettings.getServerName().toAscii().data() );
}

static int parse_commands(int argc, char*argv[])
{
	/**
	* Command line option interpreter.
	*
	* Transient command line options ( don't get remembered in settings):
	** -v | --version - display version information and exit.
	** --sqlinitialize - initialize database (requires --sqlserver & --sqldatabase).
	** -n | --nosave - don't remember any of these settings in settings file.
	** -s | --settings <settingsfile> - use this settings file.
	*
	* Persistent command line options (get remembered in settings file):
	*
	** --sqltype <type>  - currently only "mysql" suported.
	** --sqlserver <server>
	** --sqldatabase <database>
	** --sqluser <username>
	** --sqlpassword <password>
	** -p | --port <port> - port to listen on.
	** --logging <level> - 0..9 level o verbosity.	
	*/
	int rc=1;
	int n=0;
	for( n=1; n<argc && rc==1; ) {
		QString arg(argv[n]);
		if ( arg == "--version" || arg == "-v" ) {
			print_version();
			rc=-3;
		} else if ( arg == "--sqlinitialize" ) {
			serverSettings.setSqlInitialize( true );
		} else if ( arg == "--nosave" || arg == "-n" ) {
			save_settings=false;
		} else if ( arg == "--settings" || arg == "-s" ) {
			if ( n+1 < argc ) settings_file = QString(argv[++n]); else rc=-1;
		} else if ( arg == "--sqltype" ) {
			if ( n+1 < argc ) serverSettings.setSqlType(QString(argv[++n])); else rc=-1;
		} else if ( arg == "--sqlserver" ) {
			if ( n+1 < argc ) serverSettings.setSqlServer(QString(argv[++n])); else rc=-1;
		} else if ( arg == "--sqldatabase" ) {
			if ( n+1 < argc ) serverSettings.setSqlDatabase(QString(argv[++n])); else rc=-1;
		} else if ( arg == "--sqluser" ) {
			if ( n+1 < argc ) serverSettings.setSqlUser(QString(argv[++n])); else rc=-1;
		} else if ( arg == "--sqlpassword" ) {
			if ( n+1 < argc ) serverSettings.setSqlPassword(QString(argv[++n])); else rc=-1;
		} else if ( arg == "--port" || arg == "-p" ) {
			if ( n+1 < argc ) serverSettings.setPort(QString(argv[++n]).toInt()); else rc=-1;
		} else if ( arg == "--backlog" || arg == "-b" ) {
			if ( n+1 < argc ) serverSettings.setBacklog(QString(argv[++n]).toInt()); else rc=-1;
		} else if ( arg == "--logging" || arg == "-l" ) {
			if ( n+1 < argc ) serverSettings.setLogging(QString(argv[++n]).toInt()); else rc=-1;	
		} else if ( arg == "--dataroot" || arg == "-d" ) {
			if ( n+1 < argc ) serverSettings.setDataRoot(QString(argv[++n])); else rc=-1;	
		} else if ( arg == "--queryroot" || arg == "-q" ) {
			if ( n+1 < argc ) serverSettings.setQueryRoot(QString(argv[++n])); else rc=-1;	
		} else if ( arg == "--formsroot" || arg == "-f" ) {
			if ( n+1 < argc ) serverSettings.setFormsRoot(QString(argv[++n])); else rc=-1;	
		} else if ( arg == "--servername" || arg == "-S" ) {
			if ( n+1 < argc ) serverSettings.setServerName(QString(argv[++n])); else rc=-1;	
		} else if ( arg == "--help" || arg == "-h" || arg == "-?") {
			print_help();
			rc=-3;
		} else {
			rc=-2;
		}
		if ( rc >= 0 ) ++n;
	}
	switch( rc ) {
		case -1:
			fprintf( stderr, "%s: missing argument for '%s'\n", argv[0], argv[n] );
			break;
		case -2:
			fprintf( stderr, "%s: bad command '%s'\n", argv[0], argv[n] );
	        break;
	}
	return rc;
}

int main(int argc, char *argv[])
{
	int rc=0;
	serverSettings.restore();
	if ( parse_commands(argc,argv) > 0 ) {
		srand(time(0)); /** seed the random number generator */
		QCoreApplication app(argc, argv);
		if ( save_settings ) 
			serverSettings.save();
		ICEServer* server = new ICEServer;
		QObject::connect(server,SIGNAL(quit()),&app,SLOT(quit()));
		rc = app.exec();
		delete server; server=NULL;
		printf( "\n" );
	}
	return rc;
}

