/***************************************************************************
 *   Copyright (C) 2007 by Pike Aerospace Research Corporation             *
 *   info@pikeaero.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "iceserversession.h"
#include "iceserverstatistics.h"
#include "iceserversessionobjectuitransport.h"
#include "iceserversessionobjectuielementtransport.h"

#include <stdlib.h>

QMutex ICEServerSession::mRandMutex;

#define inherited ICESession

ICEServerSession::ICEServerSession ( QObject *parent, QTcpSocket* socket, ICEDatabase* database )
: inherited ( parent,true )
, mDatabase ( database )
, mCipherAttempts ( 0 )
, mPlatform ( ICEPacket::Unknown )
{
	setDebugging ( true );
	mSocket = socket;
	mUser.setLog ( &mLog );
	initialize();
	mMinuteTimer = startTimer( 1000 * 60 );
}

ICEServerSession::~ICEServerSession()
{
	for ( int n=0; n < mOpenFiles.count(); n++ )
	{
		if ( mOpenFiles[n] != NULL )
		{
			delete mOpenFiles[n];
		}
	}
}

/**
	\brief Generate a random number (integer) between "hi" and "lo".
	\param lo The low limit.
	\param hi The high limit.
 */
int ICEServerSession::generateRandomNumber ( int lo, int hi )
{
	QMutexLocker  mutexlock ( &mRandMutex );
	double flo = lo;
	double fhi = hi;
	double result = flo + ( fhi * rand() / ( RAND_MAX+flo ) );
	return (int)result;
}

/**
	\brief Generate a challenge string. The client must decode, re-code and
	transmit back, at which point we should be able to decode and compare with
	original challenge.
 */
QByteArray ICEServerSession::generateChallenge()
{
	QByteArray bytes;
	bytes.resize ( generateRandomNumber ( 200,1000 ) ); /* size of the challenge in bytes */
	for ( int i=0; i < bytes.size(); i++ )
	{
		bytes[i] = ( unsigned char ) generateRandomNumber ( 0,255 );
	}
	return bytes;
}

/**
	\brief create an asynchronous popup message on client.
	\param text The popup message text to display.
	\param code The packet return code.
	\return void
 **/
void ICEServerSession::asyncPopupMessage ( QString text, ICEPacket::PacketCode code )
{
	if ( state() != SessionClosing && state() != SessionClosed )
	{
		ICEPacket txpacket;
		txpacket.setType ( ICEPacket::AsyncPopupMessage );
		txpacket.setCode ( code );
		txpacket.setString ( text );
		tx ( txpacket );
	}
}


/**
	\brief Reply to requestDisconnet(...).
	\param rxpacket The packet containing the disconnect data.
	\param code The packet code.
	\return void
 **/
void ICEServerSession::replyDisconnect ( ICEPacket* rxpacket,ICEPacket::PacketCode code )
{
	ICEPacket txpacket;
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "reply disconnect " ) +QString::number ( code ) );
	txpacket.setType ( ICEPacket::ReplyDisconnect );
	txpacket.setCode ( code );
	setState ( SessionClosing );
	tx ( txpacket );
	QTimer::singleShot ( 3000,this,SLOT ( close() ) );
}

/**
	\brief Reply to requestSession(...).
	\param rxpacket The packet containing the session request data.
	\param code The packet code.
	\return void
 **/
void ICEServerSession::replySession ( ICEPacket* rxpacket,ICEPacket::PacketCode code )
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "reply session " ) +QString::number ( code ) );
	if ( state() == SessionChallenge )
	{
		if ( mChallenge.length() && ( rxpacket->getData() == mChallenge ) )
		{
			ICEPacket txpacket;
			setState ( SessionValidated );
			txpacket.setType ( ICEPacket::ReplySession );
			txpacket.setCode ( ICEPacket::CodeOK );
			tx ( txpacket );
			setEncrypt ( true );
			mUser.set_lastlogin( QDateTime::currentDateTime() );
		}
		else
		{
			replyDisconnect ( rxpacket,ICEPacket::CodeFailedChallenge );
		}
	}
	else
	{
		replyDisconnect ( rxpacket,ICEPacket::CodeProtocolError );
	}
}

/**
	\brief Reply to requestLogin(...).
	\param rxpacket The packet containing the login request data.
	\param code The packet code.
	\return void
 **/
void ICEServerSession::replyLogin ( ICEPacket* rxpacket,ICEPacket::PacketCode code )
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "reply login " ) +QString::number ( code ) );
	QByteArray cipher_challenge;
	if ( state() == SessionConnected )
	{
		/**
		 * Prepare a reply packet...
		 */
		ICEPacket txpacket;
		setLogin ( mUser.get_login() );
		setKey ( mUser.get_key() );
		txpacket.setType ( ICEPacket::ReplyLogin );
		txpacket.setCode ( code );
		txpacket.setLogin ( getLogin() );
		txpacket.setCipher ( rxpacket->getCipher() );
		/**
		 * Instantiate a cypher...
		 */
		QString cipherMsg = tr( "Cipher: " );
		if ( mCryptRX != NULL ) {delete mCryptRX; mCryptRX=NULL;}
		if ( mCryptTX != NULL ) {delete mCryptTX; mCryptTX=NULL;}
		switch ( rxpacket->getCipher() )
		{
			case ICEPacket::CipherSEAL:
				cipherMsg += "SEAL";
				mCryptRX = new ICECryptSEAL();
				mCryptTX = new ICECryptSEAL();
				break;
			case ICEPacket::CipherBlowfish:
				cipherMsg += "Blowfish";
				mCryptRX = new ICECryptBF();
				mCryptTX = new ICECryptBF();
				break;
			case ICEPacket::CipherLOKI91:
				cipherMsg += "LOKI91";
				mCryptRX = new ICECryptLOKI91();
				mCryptTX = new ICECryptLOKI91();
				break;
			case ICEPacket::CipherXOR:
				cipherMsg += "XOR";
				mCryptRX = new ICECrypt();
				mCryptTX = new ICECrypt();
				break;
			default:
				cipherMsg += tr( "(not supported)" );
				txpacket.setCode ( ICEPacket::CodeCipherNotSupported );
				break;
		}
		emit statusMessage ( this, ICESession::ReasonSession, cipherMsg );
		/**
		 ** Generate a challenge string...
		 */
		mChallenge = generateChallenge();
		if ( mCryptTX != NULL )
		{
			mCryptTX->cipher ( true, getKey(), &mChallenge, &cipher_challenge );
		}
		/**
		 ** Insert challenge string into packet and switch to SessionChallenge state...
		 */
		txpacket.setData ( cipher_challenge );
		if ( txpacket.getCode() == ICEPacket::CodeOK )
		{
			setState ( SessionChallenge );
			tx ( txpacket );
		}
		else
		{
			/**
			 ** ...Unless we cannot agree on the cypher, in that case, try the next one...
			 */
			if ( ++mCipherAttempts < ( int ) ICEPacket::CipherMAX )
			{
				emit statusMessage ( this,ICESession::ReasonSession,tr ( "rejected requested cipher " ) +QString::number ( rxpacket->getCipher() ) );
				setState ( SessionConnected );
				tx ( txpacket );
				setEncrypt ( false );
			}
			else
			{
				/**
				 ** If we've tried at least every cipher we know, then disconnect and reply with
				 **  ICEPacket::CodeCipherNotSupported....
				 */
				replyDisconnect ( rxpacket,ICEPacket::CodeCipherNotSupported );
			}
		}
	}
	else
	{
		emit statusMessage ( this, ICESession::ReasonSession, tr ( "session not connected" ) );
		replyDisconnect ( rxpacket,ICEPacket::CodeProtocolError );
	}
}

/**
	\brief Client has requested a session using a login ID. We
	respond with a challenge string, to which the client must
	decypher and reply.
 **/
void ICEServerSession::replyConnection ( ICEPacket* rxpacket,ICEPacket::PacketCode code )
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "reply connection (" ) +
	                     QString::number ( code ) +tr ( ") '" ) +
	                     rxpacket->getLogin() +tr ( "' protocol version " ) +
	                     QString::number ( rxpacket->getProtocolVersion() ) );
	if ( state() == SessionClosed )
	{
		mUser.load ( rxpacket->getLogin() );
		if ( mUser.valid() )
		{
			ICEPacket txpacket;
			mPlatform = rxpacket->getPlatform();
			setLogin ( mUser.get_login() );
			setKey ( mUser.get_key() );
			setState ( SessionConnected );
			txpacket.setType ( ICEPacket::ReplyConnection );
			txpacket.setCode ( code );
			txpacket.setLogin ( getLogin() );
			tx ( txpacket );
		}
		else
		{
			replyDisconnect ( rxpacket,ICEPacket::CodeUnknownLoginID );
		}
	}
	else
	{
		replyDisconnect ( rxpacket,ICEPacket::CodeProtocolError );
	}
}

/**
	\brief Client has requested a session using a new login ID. We
	respond with a challenge string, to which the client must
	decypher and reply.
 **/
void ICEServerSession::replyConnectNewLoginID ( ICEPacket* rxpacket,ICEPacket::PacketCode code )
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "reply new login id " ) +QString::number ( code ) );
	if ( state() == SessionClosed )
	{
		mUser.load ( rxpacket->getLogin() );
		if ( !mUser.valid() )
		{
			if ( rxpacket->getLogin().length() && rxpacket->getKey().length() )
			{
				emit statusMessage ( this,ICESession::ReasonSession,tr ( "create new id: '" ) +
				                     rxpacket->getLogin() +QString ( "', '" ) +
				                     rxpacket->getKey() +QString ( "'" ) );
				if ( mUser.new_id ( rxpacket->getLogin(),rxpacket->getKey() ) )
				{
					mUser.load ( rxpacket->getLogin() );
					if ( mUser.valid() )
					{
						ICEPacket txpacket;
						setLogin ( mUser.get_login() );
						setKey ( mUser.get_key() );
						setState ( SessionConnected );
						txpacket.setType ( ICEPacket::ReplyConnectNewLoginID );
						txpacket.setCode ( code );
						txpacket.setLogin ( getLogin() );
						tx ( txpacket );
						asyncPopupMessage ( tr ( "Account with ID '" ) +getLogin() +tr ( "' has been created." ) );
					}
					else
					{
						replyDisconnect ( rxpacket,ICEPacket::CodeUnknownLoginID );
					}
				}
				else
				{
					replyDisconnect ( rxpacket,ICEPacket::CodeCreateNewIDFailed );
				}
			}
			else
			{
				replyDisconnect ( rxpacket,ICEPacket::CodeInvalidCredentials );
			}
		}
		else
		{
			replyDisconnect ( rxpacket,ICEPacket::CodeDuplicateLoginID );
		}
	}
	else
	{
		replyDisconnect ( rxpacket,ICEPacket::CodeProtocolError );
	}
}

void ICEServerSession::replyDeleteLogin ( ICEPacket* rxpacket,ICEPacket::PacketCode code )
{
	fprintf ( stderr, "FIXME (do stuff) - reply delete login %d\n", code );
	if ( state() == SessionValidated )
	{
		ICEPacket txpacket;
		txpacket.setType ( ICEPacket::ReplyDeleteLogin );
		txpacket.setCode ( code );
		tx ( txpacket );
		asyncPopupMessage ( "Account with ID '"+getLogin() +"' has been deleted." );
	}
}

ICESessionObject* ICEServerSession::construct(ICESessionObject::ObjectType type,QByteArray* id,QByteArray* data)
{
	ICESessionObject* so=NULL;
	bool dataIsLocation=false;
	switch( type ) {
		case ICESessionObject::TypeUiTransport:	
			so = new ICEServerSessionObjectUiTransport(this,id); 
			break;
		case ICESessionObject::TypeUiElementTransport:	
			so = new ICEServerSessionObjectUiElementTransport(this,id); 
			break;
		default:
			break; 
	}
	if ( so != NULL ) {
		ICESessionObject* soLock = ICESessionObject::obtain( so->id(), false );
		if ( soLock != NULL ) {
			if ( data != NULL ) {
				if ( dataIsLocation ) {
					so->setLocation(QString(*data));
				} else {
					so->setUi(*data);
				}
			}
			requestSessionObjectConstruct(so);
			soLock->release(false);
		}
	}
	return so; 	
}

void ICEServerSession::replySessionObjectConstruct( ICEPacket* rxpacket,ICEPacket::PacketCode code)
{
	ICESessionObject* so=NULL;
	QByteArray id = rxpacket->getObjectId();
	switch( (ICESessionObject::ObjectType)rxpacket->getObjectType() ) {
		case ICESessionObject::TypeUiTransport: 
			so = new ICEServerSessionObjectUiTransport(this,&id);
			break;
		case ICESessionObject::TypeUiElementTransport: 
			so = new ICEServerSessionObjectUiElementTransport(this,&id);
			break;
		default:			
			break;
	}
	inherited::replySessionObjectConstruct( rxpacket, code );
}

void ICEServerSession::processReplySessionObjectUi( ICEPacket* rxpacket )
{
	ICESessionObject* so = ICESessionObject::obtain(rxpacket->getObjectId(),false);
	if ( so != NULL ) {
		if ( so->session() == this ) {
			if ( state() == ICESession::SessionValidated )
			{
				emit statusMessage ( this,ICESession::ReasonSessionObj,so->toString() + QString( "| << session object ui reply" ) );
			}
		}
		so->release(false);
	} 	
}

/**
 ** Transmmit a packet down the wire.
 **/
void ICEServerSession::tx(ICEPacket& packet)
{
	ICEServerStatistics::incrementTXPackets();
	inherited::tx(packet);
}

/**
	\brief Dispatch incoming packet to appropriate handler.
	\param rxpacket Inbound packet.
	\return void
 **/
void ICEServerSession::rx ( ICEPacket rxpacket )
{
	ICEServerStatistics::incrementRXPackets();
	switch ( rxpacket.getType() )
	{
		case ICEPacket::Empty:                       	break;
		case ICEPacket::Ping:                        	break;
		case ICEPacket::Fault:                       	break;

		case ICEPacket::RequestConnection:				replyConnection ( &rxpacket );         break;
		case ICEPacket::RequestDisconnect:				replyDisconnect ( &rxpacket );         break;
		case ICEPacket::RequestConnectNewLoginID:		replyConnectNewLoginID ( &rxpacket );  break;
		case ICEPacket::RequestLogin:					replyLogin ( &rxpacket );              break;
		case ICEPacket::RequestDeleteLogin:				replyDeleteLogin ( &rxpacket );        break;
		case ICEPacket::RequestSession:					replySession ( &rxpacket );            break;
		case ICEPacket::RequestPasswordChange:			break;

		case ICEPacket::ReplyConnection:				break;
		case ICEPacket::ReplyDisconnect:				break;
		case ICEPacket::ReplyConnectNewLoginID:			break;
		case ICEPacket::ReplyLogin:						break;
		case ICEPacket::ReplyDeleteLogin:				break;
		case ICEPacket::ReplySession:					break;
		case ICEPacket::ReplyPasswordChange:			break;
		
		case ICEPacket::ReplySessionObjectUi:			processReplySessionObjectUi( &rxpacket ); break;
		
		default:
			inherited::rx( rxpacket );
			break;
	}
}

void ICEServerSession::slotConnected()
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "session connected" ) );
}

void ICEServerSession::slotDisconnected()
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "session disconnected" ) );
	setState ( SessionClosed );
	quit();
}

void ICEServerSession::slotError ( QAbstractSocket::SocketError err )
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "session error " ) +QString::number ( err ) );
}

void ICEServerSession::slotHostFound()
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "session host found" ) );
}

void ICEServerSession::slotStateChanged ( QAbstractSocket::SocketState state )
{
	emit statusMessage ( this,ICESession::ReasonSession,tr ( "session state changed " ) +QString::number ( state ) );
}

void ICEServerSession::timerEvent(QTimerEvent* e)
{
	inherited::timerEvent(e);
	if ( e->timerId() == mMinuteTimer ) {
		mUser.set_totalminutes( mUser.get_totalminutes() + 1 );
	}
}
